/*
 *	Functions for analysing x64 architecture systems
*/

#include "../../globals.h"
#include "magic.h"
/*
 * Some alchemy:
 */

/* 
 * Function that manages mapping from linear to physical address space.
 * 
 * Good overview is available in DDI 01001.pdf (ARM documentation) at page 723+
 * (especially page 727)
 *
 *
 * priv_block_size from lin_addr are read using provided PGD (kernel PGD by default) into out_data pointer.
 * dump_offset is output parameter, if NULL, ignored.
 * Otherwise physical offset translated from linear address is stored here.
 * (!) Please note, that read block can be larger than one page. 
 * In this case offset is valid only for first part of data.
 */


int arm_read_linear(u_int64_t PGD, u_int64_t lin_addr, void *out_data, u_int64_t priv_block_size, u_int64_t *dump_offset){

	// "PGD" is needed, because each process have it's own linear address space. 
	// PGD 0 means kernel memory.
	// NOTE: Term PGD is used to refer first directory in hierarchy,

	// Find in VPT "physical" address, where data should be. (if problem, return 1)
	// Read block_size of that data into out_data (if problem, return 2)
	// return 0

	u_int64_t physical_offset = lin_addr;

	u_int64_t level1_offset = lin_addr;
	u_int64_t level2_offset = lin_addr;
	u_int64_t page_number = lin_addr;

	u_int64_t tmp_var = 0;

	dbgprint("addr: 0x%llX",lin_addr); 

	if (PGD == 0) {	//special case
		PGD = image1.VPT_address;

	/* if VPT_address is not available,
	 * we can try "best effort", and use hard-coded 
	 * "direct mapping", usually used by OS kernels.
	 * (and OS specific)
	 * for linux on arm, mapping can be found in 
	 * Documentation/arm/memory.txt
	 * parsing here:

Start		End		Use
--------------------------------------------------------------------------
ffff8000	ffffffff	copy_user_page / clear_user_page use.
				For SA11xx and Xscale, this is used to
				setup a minicache mapping.

ffff1000	ffff7fff	Reserved.
				Platforms must not use this address range.

ffff0000	ffff0fff	CPU vector page.
				The CPU vectors are mapped here if the
				CPU supports vector relocation (control
				register V bit.)

ffc00000	fffeffff	DMA memory mapping region.  Memory returned
				by the dma_alloc_xxx functions will be
				dynamically mapped here.

ff000000	ffbfffff	Reserved for future expansion of DMA
				mapping region.

VMALLOC_END	feffffff	Free for platform use, recommended.
				VMALLOC_END must be aligned to a 2MB
				boundary.

VMALLOC_START	VMALLOC_END-1	vmalloc() / ioremap() space.
				Memory returned by vmalloc/ioremap will
				be dynamically placed in this region.
				VMALLOC_START may be based upon the value
				of the high_memory variable.

PAGE_OFFSET	high_memory-1	Kernel direct-mapped RAM region.
				This maps the platforms RAM, and typically
				maps all platform RAM in a 1:1 relationship.

TASK_SIZE	PAGE_OFFSET-1	Kernel module space
				Kernel modules inserted via insmod are
				placed here using dynamic mappings.

00001000	TASK_SIZE-1	User space mappings
				Per-thread mappings are placed here via
				the mmap() system call.

00000000	00000fff	CPU vector page / null pointer trap
				CPUs which do not support vector remapping
				place their vector page here.  NULL pointer
				dereferences by both the kernel and user
				space are also caught via this mapping.


	 * so we can try to translate PAGE_OFFSET - "high_memory-1",
	 * which should be (0xC0000000 - 0xFFFFFFFF)
  	 */  
		if (image1.VPT_not_available == 1) { 

			//kernel text direct mapping
			if ((lin_addr > 0xC0000000) &&
				(lin_addr < 0xFFFFFFFF)) {
				physical_offset=lin_addr - 0xC0000000;
//				dbgprint("direct mapping 0x%llX -> 0x%llX",lin_addr,physical_offset);
			}
			// If none of mappings matches 
			if (physical_offset == lin_addr) {
//                              dbgprint("nothing matched, 0x%llX",lin_addr);
				return -1;
			}

			// Read memory
			lseek(inputfd,physical_offset,SEEK_SET); //level 0 (physical page) 
//			dbgprint("(level 0) Dirrect mapping: Reading 0x%llX from PO: 0x%llX\n",lin_addr,physical_offset);
			
			if (dump_offset != NULL) { //save physical offset to output variable
				*dump_offset=physical_offset; 
			}

			if ( (status=read(inputfd,out_data,priv_block_size)) == -1 ) {
        		        errormsg=strerror(errno);
               			fprintf(stderr, "read_liner: read() in %s: %s.\n",inputfile,errormsg);
	                	return 38;
			}
			
			return 0; //another job well done.

		}	

	}

	//VMSAv6, first level descriptors (alias PGD)

	// this part (PML4E & PDPE) is same for all page sizes in long mode
	// For extracting addresses I use bitwise and, for extracting offsets I have to shift bits.
	level1_offset = ( level1_offset & 0x00000000FFFFFFFFLLU ) >> 20; //bits ( 31:20)

	dbgprint("level1_offset: 0x%llX",level1_offset);	

	dbgprint("(level 1) PGD : 0x%llX, l1_offset: 0x%llX, pointer_size: %d",PGD,level1_offset,image1.pointer_size);	

        lseek(inputfd,PGD+level1_offset*image1.pointer_size,SEEK_SET); //level 1)
	dbgprint("(level 1) reading from : 0x%llX",PGD+level1_offset*image1.pointer_size);	

        if ( (status=read(inputfd,&page_number,image1.pointer_size)) == -1 ) {
                errormsg=strerror(errno);
                fprintf(stderr, "read_liner: read() in %s: %s.\n",inputfile,errormsg);
                return 37;
        }
	dbgprint("page_number (level 1): 0x%llX",page_number);	

	//bits ( 1:0), determines paging table structure
	tmp_var = page_number & 0x0000000000000003LLU; 
	dbgprint("bits (level 1): 0x%llX",tmp_var);	

	// Level 1 branching..

	if (tmp_var == 0) { //"fault" state
		dbgprint("Level 1 table: \"fault\"");
		return 101;
	}
	if (tmp_var == 3) { //"reserved" state (or tiny tables used)
		dbgprint("Level 1 table: \"reserved\"");
		return 102;
	}


	if (tmp_var == 1) { //"coarse" pages are used

	/* From ARM documentation:
	 * Because the upper four bits of the Page Index and low-order four bits of the 
	 * Second-level Table Index overlap, each page table entry for a large page must
	 * be repeated 16 times (in consecutive memory locations) in a coarse page table.
	 */
		dbgprint("coarse pages used for this page.");
	        level2_offset = ( level2_offset & 0x00000000000FFFFFLLU ) >> 12; // bits 19:12
	
		//bits ( 31:10), but no shift. Offset will be added later
		page_number = page_number & 0x00000000FFFFFC00LLU; 


		dbgprint("(level 2) Reading 0x%llX from PO: 0x%llX\n",lin_addr,page_number+level2_offset);
		lseek(inputfd,page_number+level2_offset,SEEK_SET); //level 2 
	        if ( (status=read(inputfd,&page_number,image1.pointer_size)) == -1 ) {
			errormsg=strerror(errno);
			fprintf(stderr, "read_liner: read() in %s: %s.\n",inputfile,errormsg);
			return 37;
		}
		dbgprint("page_number (level 2): 0x%llX",page_number);	

		//bits ( 1:0), determines paging table structure
		tmp_var = page_number & 0x0000000000000003LLU; 
		dbgprint("bits (level 1): 0x%llX",tmp_var);	

		if (tmp_var == 0) { // "falt"
			dbgprint("Level 2 table: \"fault\"");
			return 103;
		}
		if (tmp_var == 1) { //"large" (64KB) pages are used
			dbgprint("\"large\" pages used for this page.");
		
			// address of physical page
			page_number = page_number & 0x00000000FFFF0000LLU; //bits 31:16

			//offset in page, bits ( 15:0). 
			physical_offset = physical_offset & 0x000000000000FFFFLLU; 

			dbgprint("(level physical) Reading 0x%llX from PO: 0x%llX\n",lin_addr,page_number+physical_offset);
			lseek(inputfd,page_number+physical_offset,SEEK_SET); // physical level
			if (dump_offset != NULL) { //save physical offset to output variable
				*dump_offset=page_number+physical_offset;
			}

			//If we want to read more than 1 page: use recursion
		
			// offset + block_size must be less then 64KB
			if (priv_block_size > (KB_64-physical_offset)){ 
				dbgprint("arm_read_linear: page too small for reading... processing another one.");
				//recursion	
				tmp_var = KB_64-physical_offset;

				//Process rest of request recursively... beware about effectives
				arm_read_linear(PGD,lin_addr+tmp_var,out_data+tmp_var,priv_block_size-tmp_var,NULL); 
			} else{
				tmp_var=priv_block_size;
			}

			if ( (status=read(inputfd,out_data,tmp_var)) == -1 ) {
				errormsg=strerror(errno);
				fprintf(stderr, "read_liner: read() in %s: %s.\n",inputfile,errormsg);
				return 38;
			}

		}

		if ((tmp_var == 2) || (tmp_var == 3)) { 
			// "small" or "extended small" (4KB) pages are used 
			// (only if subpages are enabled, otherwise "extended small")
			// for our purpose, this is not important, offsets are same
			dbgprint("\"small\" pages used for this page.");
		
			// address of physical page
			page_number = page_number & 0x00000000FFFFF000LLU; //bits 31:12

			//offset in page, bits ( 11:0). 
			physical_offset = physical_offset & 0x0000000000000FFFLLU; 

			dbgprint("(level physical) Reading 0x%llX from PO: 0x%llX\n",lin_addr,page_number+physical_offset);
			lseek(inputfd,page_number+physical_offset,SEEK_SET); // physical level
			if (dump_offset != NULL) { //save physical offset to output variable
				*dump_offset=page_number+physical_offset;
			}

			//If we want to read more than 1 page: use recursion
		
			// offset + block_size must be less then 64KB
			if (priv_block_size > (KB_4-physical_offset)){ 
				dbgprint("arm_read_linear: page too small for reading... processing another one.");
				//recursion	
				tmp_var = KB_4-physical_offset;

				//Process rest of request recursively... beware about effectives
				arm_read_linear(PGD,lin_addr+tmp_var,out_data+tmp_var,priv_block_size-tmp_var,NULL); 
			} else{
				tmp_var=priv_block_size;
			}

			if ( (status=read(inputfd,out_data,tmp_var)) == -1 ) {
				errormsg=strerror(errno);
				fprintf(stderr, "read_liner: read() in %s: %s.\n",inputfile,errormsg);
				return 38;
			}
		} 
	}
	
	if (tmp_var == 2) { //"section" (or "supersection") pages are used
		dbgprint("section pages used for this page.");
		//TODO 16MB supersection support

		// address of physical page
		page_number = page_number & 0x00000000FFF00000LLU; //bits 31:20

		//offset in page, bits ( 19:0). 
		physical_offset = physical_offset & 0x00000000000FFFFFLLU; 

		dbgprint("(level physical) Reading 0x%llX from PO: 0x%llX\n",lin_addr,page_number+physical_offset);
		lseek(inputfd,page_number+physical_offset,SEEK_SET); // physical level
		if (dump_offset != NULL) { //save physical offset to output variable
			*dump_offset=page_number+physical_offset;
		}

		//If we want to read more than 1 page: use recursion
		
		// offset + block_size must be less then 1MB
		if (priv_block_size > (MB_1-physical_offset)){ 
			dbgprint("arm_read_linear: page too small for reading... processing another one.");
			//recursion	
			tmp_var = MB_1-physical_offset;
			//Process rest of request recursively... beware about effectives
			arm_read_linear(PGD,lin_addr+tmp_var,out_data+tmp_var,priv_block_size-tmp_var,NULL); 
		} else{
			tmp_var=priv_block_size;
		}

		if ( (status=read(inputfd,out_data,tmp_var)) == -1 ) {
			errormsg=strerror(errno);
			fprintf(stderr, "read_liner: read() in %s: %s.\n",inputfile,errormsg);
			return 38;
		}

	}

        return 0;

}


/*
 * Function for verifying if given address contains VPT 
 * (resp. PaGeDirectory) of some process.
 */

int arm_user_VPT_test(u_int64_t my_addr){

	dbgprint("0x%llX",my_addr);
	return 0;
}

/*
 * Function for verifying if given address contains system VPT 
 * (resp. PaGeDirectory)
 *
 * Verification is done by simple checking
 * if first VPT_MAGIC_ZEROES bytes contain only zeroes and 
 * if flags for first three records beyond 0xC0000000
 * (beginning of kernel address space) are same.
 * This should usually be true, but it is just heuristic, 
 * and can be wrong. XXX rewrite
 *
 * arm_VPT_test (physical address to verify )
 * return 0 on success, otherwise error > 0
 *
 */

int arm_VPT_test(u_int64_t my_addr){

	u_int64_t tmp_var;
	u_int32_t tmp32;
	int mi,ecounter=0,rcounter=0;

	// using 32bit variables. (!)
	u_int32_t tmp_small_page[ARM_VPT_MAGIC_BS/ARM_POINTER_SIZE];

	// must be 16KB aligned

	tmp_var = my_addr & 0x0000000000003FFFLLU;
	
//	dbgprint("0x%llX",my_addr);

	if (tmp_var != 0) {
		dbgprint("Not 16KB aligned!");
		return 1;
	}

	lseek(inputfd,my_addr,SEEK_SET); 

	// read whole page (16KB)
	if ( (status=read(inputfd,&tmp_small_page,ARM_VPT_MAGIC_BS)) == -1 ) {
                errormsg=strerror(errno);
                fprintf(stderr, "arm_VPT_test: read() in %s: %s.\n",inputfile,errormsg);
                return 99;
        }
	
	for (mi=0;mi < (ARM_VPT_MAGIC_BS/ARM_POINTER_SIZE);mi++) { //page
		

		// Last 2 bits determines type of page entry

		// section
		if ((tmp_small_page[mi] & 0x00000003LU) == 1) { 
			// course page entry
			// => bits 2,3,4 SBZ (should be zero)
			rcounter++;
			if ((tmp_small_page[mi] & 0x0000001CLU) != 0) {
				if (ecounter++ > ARM_VPT_MAGIC_ERR) {
//				dbgprint(" addr: 0x%llX mi=%d, ecounter: %d, rcounter: %d\n",my_addr,mi,ecounter,rcounter);
				return 2; //tolerate small amount of errors
				}
			}
		}	
		if ((tmp_small_page[mi] & 0x00000003LU) == 2) { 
			// section page entry
			// => at least bit 19 SBZ (should be zero)
			// in some modes bits 19-15, but this enough for us)
			rcounter++;
			if ((tmp_small_page[mi] & 0x00080000LU) != 0) {
				if (ecounter++ > ARM_VPT_MAGIC_ERR) {
//				dbgprint(" addr: 0x%llX mi=%d, ecounter: %d, rcounter: %d\n",my_addr,mi,ecounter,rcounter);
				return 3; //tolerate small amount of errors
				}

			}
		}
	}	


	if (rcounter < ARM_VPT_MIN_REC) {
//				dbgprint(" addr: 0x%llX mi=%d, ecounter: %d, rcounter: %d\n",my_addr,mi,ecounter,rcounter);
				return 4; // require at least few valid records
	}


	return 0;
}

/*
 * Function that tries to determine general information 
 * about analyzed image.
 *
 * for now, only system VPT is found.
 *
 * arm_determine ()
 * return 0 on success, otherwise error > 0
 * Discovered information are saved in global structure image1.
 *
 */


int arm_determine(){

	int tmp=1;
	u_int64_t addr=ARM_VPT_MAGIC_ADDR;
	
	// I realised, that I don't get too much information in 
	// this phase of analysis 
	
	printf("arm_determine\n");
	
	if (image1.VPT_user_specified == 0 ) {	// if image address was not manually specified
		while (1 != 0) {
			tmp=arm_VPT_test(addr); 
			if (tmp==0) {
				printf("VPT address found at: 0x%llX\n",addr);
			}

			addr += ARM_VPT_MAGIC_BS; // 16KB
//			dbgprint("step, addr: 0x%llX\n",addr);

			if (addr > ( ARM_VPT_MAGIC_ADDR + ARM_VPT_MAGIC_BS * ARM_VPT_MAGIC_COUNT) ) { 
				// we have tried too many pages, so something is wrong probably
				// further search will be useless
				printf("ARM VPT_address not found, terminating at: 0x%llX\n",addr);
				return 1;
			}
		}
		
		image1.VPT_address=addr; // found one! :)
		printf("VPT address found at: 0x%llX\n",image1.VPT_address);
	} else {
		if (image1.VPT_not_available == 0 ) {	// Use of PGD is not disabled	
			printf("User specified VPT address: 0x%llX\n",image1.VPT_address);
		} else {
			printf("Using hardcoded mappings instead of VPT. This may produce strange results!\n");
		}
	}

	return 0;  // 0 if all OK, otherwise error
}

