/*
 *	Functions same for analysing all architectures
*/

#include "../globals.h"
#include "i386/functions_i386.h"
#include "x86_64/functions_x86_64.h"
#include "arm/functions_arm.h"


/*
 * Some alchemy:
 */

/*
 * Function that sets architecture and OS specific 
 * settings according to values of image.arch
 *
 * user_VPT_test is used only on i386 architecture,
 * on other architectures this is same as VPT_test
 *
 * arch-os_settings ()
 * return 0 on success, otherwise error > 0
 * results are stored in global structure
 *
 */

int arch_os_settings(){

	switch(image1.arch) {
		case 1: // i386
			strncpy(image1.architecture,"i386\0",5);
			image1.read_linear=(int (*)())i386_read_linear;
			image1.VPT_test=(int (*)())i386_VPT_test;
			image1.user_VPT_test=(int (*)())i386_user_VPT_test;
			image1.determine=(int (*)())i386_determine;
			image1.pointer_size=I386_POINTER_SIZE;
			image1.integer_size=I386_INTEGER_SIZE;
			break;
		case 2: // x86_64
			strncpy(image1.architecture,"x86_64\0",7);
			image1.read_linear=(int (*)())x86_64_read_linear;
			image1.VPT_test=(int (*)())x86_64_VPT_test;
			image1.user_VPT_test=(int (*)())x86_64_VPT_test; /*_user*/
			image1.determine=(int (*)())x86_64_determine;
			image1.pointer_size=X86_64_POINTER_SIZE;
			image1.integer_size=X86_64_INTEGER_SIZE;

			break;
		case 3: // arm
			strncpy(image1.architecture,"arm\0",4);
			image1.read_linear=(int (*)())arm_read_linear;
			image1.VPT_test=(int (*)())arm_VPT_test;
			image1.user_VPT_test=(int (*)())arm_VPT_test; /*user*/
			image1.determine=(int (*)())arm_determine;
			image1.pointer_size=ARM_POINTER_SIZE;
			image1.integer_size=ARM_INTEGER_SIZE;

			break;
	}

	return 0;
}

/*
 * Function with heuristic for architecture guessing
 *
 * This function is not really useful.
 * Original idea was that there may exists some 
 * heuristic, that can identify different types 
 * of architectures. I was not able to find such heuristic.
 * Because number of different architectures is small,
 * best solution seem to be try then one by one.
 * Usually, this is not needed, because user 
 * knows architecture of dumped system. (or have a hint) 
 *
 *
 * guess_architecture ()
 * return 0 on success, otherwise error > 0
 * results are stored in global structure
 *
 */

int guess_architecture(){

	printf("Guessing architecture...\n");
	
	if (image1.arch==0) {
		printf("Architecture not provided...using i386.\n");
		image1.arch=1; //i386
	}; 


	return 0;
}

/*
 * Function that goes trough simple list and counts its size
 *
 * Simple list look like this:
 *
 *     entry    entry    entry     entry
 *       |    --> |    --> |    --> ...
 *   off |   --   |   --   |   --      
 *   set |  --    |  --    |  --       
 *       |---     |---     |---        
 *
 * offset is distance between address where pointer points
 * and address of next pointer 
 *
 * simple_list_size(address , offset in structure)
 * return size of list on success, otherwise error < 0
 *
 */

int simple_list_size(u_int64_t addr,int offset){

	u_int64_t tmp1=0,tmp2=0;
	int size=0;

//	dbgprint("addr: 0x%llX, offset %d",addr,offset);
	for (tmp1=addr;;size++){
//		dbgprint("ptr: 0x%llX, size %d",tmp1,size);
	
		if (size > MAX_LIST_SIZE ) {
			return -2; //List too long
		} 	

		if ( (status=image1.read_linear(0,tmp1+offset,&tmp2,image1.pointer_size,NULL)) != 0 ) {
			errormsg=strerror(errno);
			dbgprint("read() in %s: %s.\n",inputfile,errormsg);
			return -1; //List should be ended by NULL pointer.
		}
	
		if ((tmp1 == tmp2 ) || (tmp2 == 0 )) {
			return (size+1); 
			// It is strange, but sometimes list 
			// seems to be finished by pointer to itself.
		}
		tmp1=tmp2;
	}

}

/*
 * Function for verifying if given address contains structure list_head.
 *
 * Verification is done by finding and checking double linked 
 * "list_head". This way is chosen instead of classic 
 * pattern matching, because I believe it is more flexible and 
 * can be used for almost every 2.6 kernel version. (even future;)
 * It does not depend on task_struct definition too much, only 
 * important characteristic is number of list_heads between 
 * "comm" array and list_head "tasks". This is specified by
 * TS_MAGIC_HEADLIST_NUMBER. 
 *
 * Correct list_head can be verified easily:
 * Data at address point to (virtual) address where is stored another
 * list_head. Second pointer in this second head list point back
 * to (virtual) address where we started.
 *
 *  ...   |--------> |---------->  |---------->   ...
 *          <------  | <-----------| <-------- 
 *
 * Note: This will not find processes, that are inactive,
 * has already ended in past, or something similar. 
 * Therefore, if you want to discover all the processes, 
 * you have to generate pattern and then use bgrep utility.
 *
 *
 *  verify_cyclic_list_head (address to verify )
 *  return 0 on success, otherwise error > 0
 *
 * */

int verify_cyclic_list_head(u_int64_t ver_addr){

	u_int64_t tmp_addr1=0,tmp_addr2=0,tmp_addr3=0;

//	dbgprint("start, addr: %llX",ver_addr); 
	if ( (status=lseek(inputfd,ver_addr,SEEK_SET)) == -1 ) {
		errormsg = strerror(errno);
                fprintf(stderr, "verify_cyclic_list_head: lseek() to 0x%llX failed because of %s.\n",ver_addr,errormsg);
                return(1);
        }

	if ( (status=read(inputfd,&tmp_addr1,image1.pointer_size)) == -1 ) {
                errormsg=strerror(errno);
                fprintf(stderr, "verify_cyclic_list_head: read() in %s: %s.\n",inputfile,errormsg);
                return(2);
        }

	if (tmp_addr1 == 0) {
//		printf("verify_cyclic_list_head: Zero.\n");
		return 6;
	}

	if ( (status = image1.read_linear(0,tmp_addr1,&tmp_addr2,image1.pointer_size,NULL) == -1 )) {
                fprintf(stderr,"verify_cyclic_list_head: 0x%llX seems is not valid virtual address.\n",tmp_addr1);
                return(3);
        }
	
	if ( (status = image1.read_linear(0,tmp_addr2+image1.pointer_size,&tmp_addr3,image1.pointer_size,NULL) == -1 )) {
                fprintf(stderr,"verify_cyclic_list_head: 0x%llX seems is not valid virtual address.\n",tmp_addr3);
                return(4);
        }
	
	if (tmp_addr1 != tmp_addr3) {
//		dbgprint("verify_cyclic_list_head: addresses 0x%llX and 0x%llX does not match.\n",tmp_addr1,tmp_addr3);
		return(5);
	}
	
	// We need to verify second pointer of list_head too, there were some false positives
	// same algorithm, except that "prev" pointer points to 
	// (address - image1.pointer_size) of reverse pointer.

	if ( (status=lseek(inputfd,ver_addr-image1.pointer_size,SEEK_SET)) == -1 ) {
		errormsg = strerror(errno);
                fprintf(stderr, "verify_cyclic_list_head: 2: lseek() to 0x%llX failed because of %s.\n",ver_addr,errormsg);
                return(6);
        }

	if ( (status=read(inputfd,&tmp_addr1,image1.pointer_size)) == -1 ) {
                errormsg=strerror(errno);
                fprintf(stderr, "verify_cyclic_list_head: 2: read() in %s: %s.\n",inputfile,errormsg);
                return(7);
        }

	if (tmp_addr1 == 0) {
//		printf("verify_cyclic_list_head: Zero.\n");
		return 8;
	}


	if ( (status = image1.read_linear(0,tmp_addr1+image1.pointer_size,&tmp_addr2,image1.pointer_size,NULL) == -1 )) {
                fprintf(stderr,"verify_cyclic_list_head: 2: 0x%llX seems is not valid virtual address.\n",tmp_addr1);
                return(9);
        }
	

	if ( (status = image1.read_linear(0,tmp_addr2,&tmp_addr3,image1.pointer_size,NULL) == -1 )) {
                fprintf(stderr,"verify_cyclic_list_head: 2: 0x%llX seems is not valid virtual address.\n",tmp_addr3);
                return(10);
        }
	
	if (tmp_addr1 != tmp_addr3) {
//		printf("verify_cyclic_list_head: 2: addresses 0x%llX and 0x%llX does not match.\n",tmp_addr1,tmp_addr3);
		return(11);
	}

	return 0;
}



/*
 *  lin_verify_vfsmount (pointer to vfsmount structure)
 *  return 0 on success, otherwise -1
 *
 *  vfsmount structure is "recognized" by simple test:
 *  should contain pointer to string "/dev/".
 *  (char *mnt_name)
 *  This is alone not a strong heuristic, but in combination
 *  with specific chain where structure should be found when trying 
 *  to list open files, it is enough.
 */

int lin_verify_vfsmount(u_int64_t ver_addr) {

	int tmp,tmp2;
	u_int64_t tmp3=0;
	char pattern[VFS_MAGIC_PATTERN_LEN];

	// Adding another check for vfsmount,
	// first member of vfsmount struct should be list_head mnt_hash.
	// TODO change, to something more reliable
	if ((ver_addr != 0xffffffff) && (ver_addr != 0)) { //just optimization
		tmp=verify_cyclic_list_head(ver_addr);
		if ( tmp != 0) { 
			return -2;
		}	
		printf("lin_verify_vfsmount: found HL at %llX\n",ver_addr);
	}
	

	for (tmp=1;tmp<VFS_MAGIC_MAX_TRIES;tmp++){
//		dbgprint("	lin_verify_vfsmount: %d: %llX ::\n",tmp,ver_addr+(tmp*image1.pointer_size)); //DBG
		if ( (status = image1.read_linear(0,ver_addr+(tmp*image1.pointer_size),&tmp3,image1.pointer_size,NULL) == -1 )) { //read pointer to string
//               	dbgprint("	lin_verify_vfsmount: 0x%llX seems is not valid virtual address.\n",ver_addr+(tmp*image1.pointer_size));
                //	return -3;
       		} else {
			if ( (status = image1.read_linear(0,tmp3,pattern,VFS_MAGIC_PATTERN_LEN,NULL) == -1 )) { //read string
//	      	        	dbgprint("	lin_verify_vfsmount: 0x%llX seems is not valid virtual address.\n",tmp3);
	        	  //      return -4;
	        	} else {
				tmp2=strncmp(pattern,VFS_MAGIC_PATTERN,VFS_MAGIC_PATTERN_LEN); //compare found string
				if (tmp2 == 0) {	// found "/dev/"
					printf("	lin_verify_vfsmount: Found some \"/dev/\" at %llX (offset %d) of vfsmount 0x%llX, dbg: 0x%llX\n",tmp3,tmp*image1.pointer_size,ver_addr,ver_addr+(tmp*image1.pointer_size)); //DBG
					return 0;  // Found => It is vfsmount
					}		
				}
			}
		}

	return -1;	// characteristic string not found => prob. not vfsmount
}



/*
 *  lin_verify_qstr (pointer to "qstr" structure )
 *  return 0 on success, otherwise error
 *
 *  Used for listing of open files per process.
 *    d_name -> name
 *    (qstr)   (char)
 *
 *  This function verifies, if pointer points to qstr structure.
 *  qstr is elemental structure, and we assume, that will not change (there is no reason)
 *  Verification is done by checking, if string that "name" points to is really "len"
 *  bytes long (looking for \0) and "len" is not 0.
 *
 *  This is qstr definition:
 *
 *  struct qstr {
 *          unsigned int hash;
 *          unsigned int len;
 *          const unsigned char *name;
 *  };
 *                          
 *
 */

int lin_verify_qstr(u_int64_t ver_addr) {
	
	u_int64_t tmp1=0,tmp2=0;
//	size_t tmp3;
	char tmp_filename[QS_MAGIC_MAX_FILENAME_SIZE];

	if ( (status = image1.read_linear(0,ver_addr+image1.integer_size,&tmp1,image1.integer_size,NULL) == -1 )) { //skip "hash", read string length
                printf("lin_verify_qstr: 0x%llX seems is not valid virtual address.\n",ver_addr+image1.integer_size);
                return(2);
        }	

	if ( (status = image1.read_linear(0,ver_addr+(image1.integer_size*2),&tmp2,image1.pointer_size,NULL) == -1 )) { //read pointer to string
                printf("lin_verify_qstr: 0x%llX seems is not valid virtual address.\n",ver_addr+image1.integer_size);
                return(3);
        }

	if ((tmp1 > QS_MAGIC_MAX_FILENAME_SIZE) || (tmp1 == 0)) { //this should be length of string
		dbgprint("length too big or small: %llX.\n",tmp1);
		return -2;
	}

	if ( (status = image1.read_linear(0,tmp2,tmp_filename,tmp1,NULL) == -1 )) { //read pointer to string
                printf("lin_verify_qstr: 0x%llX seems is not valid virtual address.\n",ver_addr+image1.integer_size);
                return(4);
        }

	if ( strnlen(tmp_filename,tmp1+1) != (tmp1)) {
//		dbgprint("sizes does not match: %llX vs %d, %d.\n",tmp1,strlen(tmp_filename),tmp2);
		return -1;
	}


	//try to read whole string
	//if size is ok, return 0

	printf("found filename: !!!%s!!! with %d at %llX (tmp1: %llu)\n",tmp_filename,strlen(tmp_filename),tmp2,tmp1);// DBG

	return 0;
}

/*
 *  lin_verify_dentry (pointer to "dentry" structure )
 *  return 0 on success, otherwise -1
 *  determined distances are saved into image1
 *
 *  Used for listing of open files per process.
 *  Chain we are following looks like:
 *
 *  => task_struct ->       files   ->      fdt     ->      fd      ->      f_path  ->      dentry  ->      d_name-> name
 *  		    (files_struct)      (fdtable)        (file **)          (path)          (dentry)        (qstr)  (char)
 *
 *  This function determines distances in structure dentry
 *  It is better to determine structures by more "sub" heuristics, because amount of tries needed is decreased.
 *  (and amount of false positives too)
 *  
 */

int lin_verify_dentry(u_int64_t ver_addr) {

	int tmp,tmp2;

	for (tmp=1;tmp<DT_MAGIC_MAX_TRIES;tmp++){
	
		dbgprint("lin_verify_dentry: %d: %llX ::\n",tmp,ver_addr+(tmp*image1.pointer_size)); //DBG
		tmp2=lin_verify_qstr(ver_addr+(tmp*image1.pointer_size));
		
		if (tmp2 == 0) {	// found qstr
			
			printf("lin_verify_dentry: %d: %llX ::\n",tmp,ver_addr+(tmp*image1.pointer_size));	
			dbgprint("	lin_verify_dentry: Found some qstr at offset %d\n",tmp*image1.pointer_size); //DBG

			image1.dt_name_dst=tmp*image1.pointer_size; //save determined distance

				return 0;  // Found => It is task_list
			}		
		}

	return -1;	// qstr not found
}

/*
 *  lin_determine_up_to_path (pointer to "tasks" list_head in task_struct )
 *  return 0 on success, otherwise -1
 *  determined distances are saved into image1
 *
 *  function that tries to determine distances in structures needed
 *  for listing of open files per process up to f_path structure.
 *  Chain we are following looks like:
 *
 *  => task_struct ->       files   ->      fdt     ->      fd      ->      f_path  ->      dentry  ->      d_name-> name
 *  		    (files_struct)      (fdtable)        (file **)          (path)          (dentry)        (qstr)  (char)
 *
 *  So this function determines distances in structures files_struct, fdtable, file and path.
 *  It is better to determine structures by more "sub" heuristics, because amount of tries needed is decreased.
 *  (and amount of false positives too)
 *  
 */

int lin_determine_up_to_path(u_int64_t ver_addr) {

	int tmp,tmp2,tmp3,tmp4,tmp5;
	u_int64_t tmpl1=0,tmpl2=0,tmpl3=0,tmpl4=0;

	for (tmp=1;tmp<TS_MAGIC_FILES_MAX_TRIES;tmp++){	//---Looking for "files" structure
		dbgprint("tmp: %d, ver_addr+(tmp*image1.pointer_size): %llX\n",tmp,ver_addr+(tmp*image1.pointer_size)); //DBG
		if ( (status = image1.read_linear(0,ver_addr+(tmp*image1.pointer_size),&tmpl1,image1.pointer_size,NULL) == -1 )) { //read pointer to "files"
                	printf("	lin_determine_up_to_path: 0x%llX seems is not valid virtual address.\n",ver_addr+(tmp*image1.pointer_size));
       		 } else {

			for (tmp2=1;tmp2<FS_MAGIC_MAX_TRIES;tmp2++){	//---Looking for "fdt" structure
//				dbgprint("2, tmp2: %d,tmpl1+(tmp2*image1.pointer_size): %llX \n",tmp2,tmpl1+(tmp2*image1.pointer_size)); //DBG
				if ( (status = image1.read_linear(0,tmpl1+(tmp2*image1.pointer_size),&tmpl2,image1.pointer_size,NULL) == -1 )) { //read pointer to string
		                	printf("	lin_determine_up_to_path: 0x%llX seems is not valid virtual address.\n",tmpl1+(tmp2*image1.pointer_size));
		       		 } else {

					for (tmp3=1;tmp3<FD_MAGIC_MAX_TRIES;tmp3++){	//---Looking for "fd" structure
//						dbgprint("3, tmp3: %d,tmpl2+(tmp3*image1.pointer_size): %llX \n",tmp3,tmpl2+(tmp3*image1.pointer_size)); //DBG
						if ( (status = image1.read_linear(0,tmpl2+(tmp3*image1.pointer_size),&tmpl3,image1.pointer_size,NULL) == -1 )) { //read pointer to string
				                	printf("	lin_determine_up_to_path: 0x%llX seems is not valid virtual address.\n",tmpl2+(tmp3*image1.pointer_size));
				       		 } else {

							for (tmp4=1;tmp4<FL_MAGIC_MAX_TRIES;tmp4++){	//---Looking for "f_path" structure
//								dbgprint("4, tmp4: %d, tmpl3+(tmp4*image1.pointer_size): %llX ::\n",tmp4,tmpl3+(tmp4*image1.pointer_size)); //DBG
								if ( (status = image1.read_linear(0,tmpl3+(tmp4*image1.pointer_size),&tmpl4,image1.pointer_size,NULL) == -1 )) { //read pointer to string
						                	printf("	lin_determine_up_to_path: 0x%llX seems is not valid virtual address.\n",tmpl3+(tmp4*image1.pointer_size));
						       		 } else {
			
									tmp5=lin_verify_vfsmount(tmpl4); // Now call verification function (vfsmount is 1st member of fpath)	
									if (tmp5 == 0) {	// found !
										printf("Found some path struct at offset %llX, (via vfsmount struct at %llX)\n",tmpl3,tmpl4); //DBG
										image1.ts_hl_fs_dst=tmp*image1.pointer_size;       //Distance from "tasks" list_head to "files" files_struct in task_struct
										image1.fdt_fd_dst=tmp2*image1.pointer_size;       //Distance from beginning of structure to "fdt" fdtable in files_struct
										image1.fl_fpath_dst=tmp3*image1.pointer_size;       //Distance from beginning of structure to "f_path" path in file structure
										image1.fs_vfs_dst=tmp4*image1.pointer_size;       //Distance from beginning of structure to vfsmount in fs_struct	
										image1.one_of_paths=tmpl3;		//Linear pointer to one of path structures
										// image1.ts_hl_fs_dst=tmp*image1.pointer_size;       //Distance from "tasks" list_head to "files" files_struct in task_struct
										// image1.fdt_fd_dst=tmp2*image1.pointer_size;       //Distance from beginning of structure to "fdt" fdtable in files_struct
										// image1.fl_fpath_dst=tmp3*image1.pointer_size;       //Distance from beginning of structure to "f_path" path in file structure
										// image1.fs_vfs_dst=tmp4*image1.pointer_size;       //Distance from beginning of structure to vfsmount in fs_struct	

										return 0;  // Found => It is task_list
									}
								}
							}
						}
					}
				}
			}
		}

	}

	return -1;	// Not found
}

/*
 *  determine_open_files (pointer to "tasks" list_head in task_struct )
 *  return 0 on success, otherwise -1
 *
 *  function that tries to determine distances in all structures needed
 *  for listing of open files per process. 
 *  This will definitely not work in 2.4 kernels.
 */

int lin_determine_open_files(u_int64_t ver_addr) {

	int tmp,tmp2;
	
	printf(" determine_open_files: 0x%llX.\n",ver_addr);
	tmp=lin_determine_up_to_path(ver_addr+image1.ts_hl_comm_dst);	//"files" is after "comm" field

	if (tmp != 0) {
		printf(" determine_open_files: lin_determine_up_to_path  on 0x%llX (0x%llX) returned  %d.\n",ver_addr,ver_addr+image1.ts_hl_comm_dst,tmp);
		return -2;
	}
	dbgprint("determine_up_to_path complete\n"); //dbg

	for (tmp=0;tmp<FL_MAGIC_MAX_TRIES;tmp++){	//---Looking for "fd" structure
		printf(" %d: %llX ::\n",tmp,image1.one_of_paths+(tmp*image1.pointer_size)); //DBG
		tmp2=lin_verify_dentry(image1.one_of_paths+(tmp*image1.pointer_size)); // Now call verification function	
	
		if (tmp2 == 0) {	// found !
			printf("	Found some dentry at offset %llX\n",image1.one_of_paths+(tmp*image1.pointer_size)); //DBG
			image1.pt_dentry_dst=tmp*image1.pointer_size;	//save offset inside path structure (to dentry)
			printf("Determination for listing of open files completed.\n");			
			return 0;  // Found => It is task_list	
		}
	}
		
	return -1;
}

/*
 *  function that tries to determine distances in all structures needed
 *  for listing of open files per process. 
 *  Should work in most BSD kernels.
 *
 *  => struct proc ->     p_fd     ->   fd_ofiles   ->   TODO
 *  		      (filedesc*)        (file**)        ()
 *

 *  bsd_determine_open_files (pointer to "proc" list_head in proc struct )
 *  return 0 on success, otherwise -1
 
 */

int bsd_determine_open_files(u_int64_t ver_addr) {

	// TODO?
	// there is no standard lsof utility in bsd.
	// listing file names seems to be quite complicated this way.

	dbgprint("NOT IMPLEMENTED");
		
	return -1;
}

/*
 *  verify_task_struct (physical address of string with process name inside process structure )
 *  
 *  Function verify, if there are some list_heads (cyclic or bsd style)
 *  around given address. List_heads must be long enough.
 *
 *  Works for multiple operating systems, (lin, win, bsd)
 *
 *  return 0 on success, otherwise negative value
 */

int verify_task_struct(u_int64_t ver_addr) {

	int tmp,tmp2,tmp3=TS_MAGIC_HEADLIST_MAX,tmp5=-1,tmp6,tmp7;
	u_int64_t tmp4=0,bsd_list_start=0,bsd_list_start_tmp=0;
	int found_os=0,bsd_list_strings_size=0,max_bsd_list_strings_size=0;

//	dbgprint("0x%llX\n",ver_addr); //DBG

	for (tmp=1;tmp<TS_MAGIC_MAX_TRIES;tmp++){
	
//		dbgprint(" %d: %llX",tmp,ver_addr-(tmp*image1.pointer_size)); //DBG
		tmp2=verify_cyclic_list_head(ver_addr-(tmp*image1.pointer_size));
		
		if (tmp2 == 0) {	// found some list_head
		
//			dbgprint("Found some list_head at negative offset %d",tmp*image1.pointer_size); //DBG

			image1.ts_headlists[++tmp5]=ver_addr-(tmp*image1.pointer_size); //save each list_head
//			dbgprint("headlists[%d]:%llX \n",tmp5,image1.ts_headlists[tmp5]);

			tmp++; 
			/* because of there can be (and are) 3 head lists in row, 
			   and without this
			   our algorithm would find 5 "list_heads" */

			if (--tmp3 == 0 ) { //maximum list_head's found
					    // this should not happen
//			dbgprint("maximum list_head's found: %d",TS_MAGIC_HEADLIST_MAX);
				break; // enough head list found
			}		
		}
	}

//	dbgprint("listheads found: %d, needed: %d\n",tmp5, TS_MAGIC_HEADLIST_NUMBER);
	
	if (tmp5 >= TS_MAGIC_HEADLIST_NUMBER) { //If this is really task_struct
		tmp2=tmp5-TS_MAGIC_HEADLIST_FRONT_NUMBER;
//		debug();
//		dbgprint("tmp2: %d\n",tmp2);

		//This step is needed, because verify_cyclic_list_head seems to produce 
		// some false positives, so TS_MAGIC_HEADLIST_FRONT_NUMBER may be not enough accurate.
		tmp3=0; tmp6=0; tmp7=0;
		for (tmp=max(tmp2-TS_MAGIC_HEADLIST_SPREAD,0);tmp <= min(tmp2+TS_MAGIC_HEADLIST_SPREAD,tmp5);tmp++,tmp3=0,tmp7=0){ 

//			dbgprint("listing: 0x%llX",image1.ts_headlists[tmp]);
			// BSD style list_head
			if ((image1.os==OS_BSD)|| (image1.os==OS_UNKNOWN)) {
				tmp7=bsd_list_head_size(image1.ts_headlists[tmp],&bsd_list_start_tmp,ver_addr-image1.ts_headlists[tmp]+image1.pointer_size,&bsd_list_strings_size);
//				dbgprint("BSD list string size: %d",bsd_list_strings_size);
			} 
			// Lin/Win style list_head 
			if ((image1.os < OS_CIRC_HL)|| (image1.os==OS_UNKNOWN)) {
				tmp3=lin_list_head_size(image1.ts_headlists[tmp]);
			}

			if (max(tmp3,tmp7) >= tmp6) {// more and not loop (loop < 0)

	/*  Sometimes, there are multiple pointers to list_head we are looking for.
	    Therefore our used offset can be wrong, and we should count string for
	    each element of list. */

//				dbgprint("Saving local maximum:(1.try) 0x%llX",image1.ts_headlists[tmp]);

				if ((max(tmp3,tmp7) > tmp6) || (bsd_list_strings_size > max_bsd_list_strings_size)) { 
//					dbgprint("Saving local maximum: 0x%llX, circ_hl: %d, bsd_hl: %d",image1.ts_headlists[tmp],tmp3,tmp7);
					image1.ts_headlists_number=tmp; //save which list_head is right one
					if (tmp3 >= tmp7) {
						tmp6=tmp3;		// remember local maximum
						found_os=OS_CIRC_HL;	// list_head type
					} else {
						tmp6=tmp7;		// remember local maximum
						found_os=OS_BSD;	// list_head type
						bsd_list_start=bsd_list_start_tmp; //remember first entry of list
						max_bsd_list_strings_size=bsd_list_strings_size;
					}
				}
			}
		}
			
		// Additional check, if found list_head is not too short
		if ( tmp6 < TS_MAGIC_MIN_LIST_SIZE) {
			dbgprint("suspicious struct at %llX, but too short list (%d).\n",ver_addr,tmp6);
//			fprintf(stderr,"	verify_task_struct: suspicious struct at %llX, but too short list (%d).\n",ver_addr,tmp6);
			return -8;
		}

		// So at this point we know position of right list_head

		printf("	verify_task_struct: List size %d\n",tmp6);
//		dbgprint("Found tasks at negative offset %d\n",tmp*image1.pointer_size); //DBG

		// set distance from list_head to process name string for this kernel.
		image1.ts_hl_comm_dst=ver_addr-image1.ts_headlists[image1.ts_headlists_number]+image1.pointer_size;
		
		if ( (status=lseek(inputfd,image1.ts_headlists[image1.ts_headlists_number],SEEK_SET)) == -1 ) {
	                errormsg = strerror(errno);
	                fprintf(stderr, "	verify_task_struct: lseek() to 0x%llX failed because of %s.\n",ver_addr,errormsg);
	                return(-2);
	        }

	        if ( (status=read(inputfd,&tmp4,image1.pointer_size)) == -1 ) {
	                errormsg=strerror(errno);
	                fprintf(stderr, "	verify_cyclic_list_head: read() in %s: %s.\n",inputfile,errormsg);
	                return(-3);
	        }
		
//		dbgprint("Adding process at 0x%llX to image1.ts_one_of_lists.",tmp4);	
		
		if (image1.os==OS_UNKNOWN) { // We found list of some type, so set OS if unknown
			image1.os=found_os;
		}

		// special case, bsd list is not circular
		// so we need to point to the first entry
		if ((image1.os==OS_BSD) && (bsd_list_start != 0)) { tmp4 = bsd_list_start;}
		image1.ts_one_of_lists=tmp4;
		return 0;  // Found => It is task_list
	}
	//otherwise it is not task_struct
	return -1;
}

/*
 *
 *  On linux, interesting fields in "module" are just near. This should not change 
 *  (no reason), so this function is much more simple then verify_task_struct.
 *  BSD system seems to use only few (if even some) modules at runtime,
 *  so this type of heuristic would not work anyway.
 *  More simplicity means better reliability and failure resistance.
 *  Modify this function, if you are dealing with some exotic OS.
 *
 *  lin_verify_module (pointer to "name" field of module )
 *  return distance between name and list_head in bytes on success, otherwise < 0
 *
 *  
 */

int lin_verify_module(u_int64_t ver_addr) {

	int tmp2,tmp3;
	u_int64_t tmp4=0;

//	dbgprint("verify_module: %llX: %llX ::\n",ver_addr,ver_addr-(1*image1.pointer_size)); 
	tmp2=verify_cyclic_list_head(ver_addr-(1*image1.pointer_size)); 
		
	if (tmp2 == 0) {	// found some list_head
//		dbgprint("	verify_module: Found some list_head at negative offset %d\n",1*image1.pointer_size); 
		image1.md_headlists[0]=ver_addr-(1*image1.pointer_size); //save list_head
//		dbgprint("	verify_module: headlists[%d]:%llX \n",tmp2,image1.ts_headlists[0]);
	
		 //If this is really module

		tmp3=lin_list_head_size(image1.ts_headlists[0]);
//		dbgprint("        verify_module: tmp3 (size): %d of %llX\n",tmp3,image1.md_headlists[0]);
		if (tmp3 > 0) {// more and not loop (loop < 0)
			image1.md_headlists_number=0; //save which list_head is right one
		}
		

		// So at this point we know position of right list_head

		printf("	lin_verify_module: Found module struct at negative offset %d\n",1*image1.pointer_size+image1.integer_size); // + integer size because of enum
		image1.md_hl_name_dst=ver_addr-image1.md_headlists[0]+image1.pointer_size; //set distance from headlist to name for this kernel.
		
		if ( (status=lseek(inputfd,image1.md_headlists[0],SEEK_SET)) == -1 ) {
	                errormsg = strerror(errno);
	                fprintf(stderr, "lin_verify_module: lseek() to 0x%llX failed because of %s.\n",ver_addr,errormsg);
	                return(-2);
	        }

	        if ( (status=read(inputfd,&tmp4,image1.pointer_size)) == -1 ) {
	                errormsg=strerror(errno);
	                fprintf(stderr, "lin_verify_module: read() in %s: %s.\n",inputfile,errormsg);
	                return(-3);
	        }
//		dbgprint("lin_verify_module: image1.md_one_of_lists: 0x%X",tmp4);	
		image1.md_one_of_lists=tmp4; //hm, dirty trick, this should be in determine module
		return 0;  // Found => It is module
	}
	//otherwise it is not module
	return -1;
}


/* 
 * Function for finding VPT of given process 
 * (resp. PaGeDirectory).
 *
 * This function is assumed to be called only once.
 * When distances used in analyzed image are known,
 * it is faster and safer to use them directly.
 * For each architecture appropriate heuristic function is called.
 *
 * lin_find_process_VPT (linear address of "tasks" list_head)
 * return 0 on success, otherwise error > 0
 * Discovered distances are saved in global structure image1.
 *
 * NOTE: ts_hl_mm_dst distance must be known before calling this function.
 * 
 */

int lin_find_process_VPT(u_int64_t my_addr){

	int mi=0;
	u_int64_t tmp2=0,tmp3=0;

	//read mm_struct address
	if ( (status=image1.read_linear(0,my_addr+image1.ts_hl_mm_dst,&tmp2,image1.pointer_size,NULL)) != 0 ) { 
		errormsg=strerror(errno);
		fprintf(stderr, "lin_find_process_VPT: read() in %s: %s.\n",inputfile,errormsg);
		return -2; //bad ptr
	}

	//guessing "mm_pgd_dst"
	for (mi=0;mi < MM_MAGIC_PGD_DST_P*image1.pointer_size;mi+=image1.pointer_size) {	
		if ( (status=image1.read_linear(0,tmp2+mi,&tmp3,image1.pointer_size,NULL)) != 0 ) {
			errormsg=strerror(errno);
			fprintf(stderr, "lin_find_process_VPT: read() in %s: %s.\n",inputfile,errormsg);
		}

		// On all supported architectures, PGD address is aligned at least 
		// on 1KB boundary. This simple check increases reliability.
		if ( (tmp3 & 0xFFF) == 0  ) {
			//if this pointer points to VPT (resp. PGD)	
			if ( image1.user_VPT_test(tmp3) == 0) {
				image1.mm_pgd_dst=mi; //then set distance we found (inside mm_struct from start to pgd)
				printf("Found userspace PGD at 0x%llX and determined distances: mm_pgd: %d\n",tmp3,mi);
				return 0;
			}
		}
	}

	return 1; //did not found VPT(resp. PGD)
}


/*
 * Function for finding VMAs of given process 
 *
 * This function is assumed to be called only once,
 * after "ts_hl_mm_dst" is known.
 * When distances used in analyzed image are known,
 * it is faster and safer to use them directly.
 *
 * VMAs are parts of process mapped memory,
 * we need their addresses to perform full dump
 * of process memory.
 * 
 * Function tries address around values that was used in 
 * available linux 2.6 kernel, and tries, if they match definition.
 * Success is not guaranteed, but it was working nice even for kernels
 * with random debugging symbols compiled in.
 *
 * lin_find_process_VMA (linear address of "tasks" list_head)
 * return 0 on success, otherwise error > 0
 * Discovered distances are saved in global structure image1.
 *
 * */

int lin_find_process_VMA(u_int64_t my_addr){

	u_int64_t tmp2=0,tmp3=0,tmp4=0;
	int tmptmp=0,max=0;

//	dbgprint("address: 0x%llX",my_addr);

	for (;tmptmp<(TS_MAGIC_HL_MM_DST_P*image1.pointer_size);tmptmp+=image1.pointer_size) {
	
//		dbgprint("reading from 0x%llX, offset: %d",my_addr+tmptmp,tmptmp);
		
		if ( (status=image1.read_linear(0,my_addr+/*image1.ts_hl_mm_dst*/tmptmp,&tmp2,image1.pointer_size,NULL)) != 0 ) { 
			//first read pointer to mm_struct
			errormsg=strerror(errno);
		        dbgprint("read() in %s: %s.\n",inputfile,errormsg);
		}
	
//		dbgprint("read 1: 0x%llX",tmp2);
	
		if ( (status=image1.read_linear(0,tmp2,&tmp3,image1.pointer_size,NULL)) != 0 ) { 
			//read first member of mm_struct structure
			//this should be vm_area_struct *
			//If we want to be more generic, (other OSes)
			//offset determination should be performed.
			errormsg=strerror(errno);
		        dbgprint("read() in %s: %s.\n",inputfile,errormsg);
		}
//		dbgprint("read 2: 0x%llX",tmp3);

		if ( (status=image1.read_linear(0,tmp3,&tmp4,image1.pointer_size,NULL)) != 0 ) { 
			//read first member of vm_area_struct structure
			//this should be mm_struct *, we belong to
			errormsg=strerror(errno);
		        dbgprint("read() in %s: %s.\n",inputfile,errormsg);
		}
//		dbgprint("read 3: 0x%llX",tmp4);

		/* now we have in tmp2 pointer to vm_area_struct
		 * should look like this:
		 *
		 *struct vm_area_struct {
		 *  struct mm_struct * vm_mm;       // The address space we belong to. 
		 *  unsigned long vm_start;         // Our start address within vm_mm.
		 *  unsigned long vm_end;           // The first byte after our end address within vm_mm. 
		 * // linked list of VM areas per task, sorted by address 
		 * struct vm_area_struct *vm_next;
		 *
		 * tmp3 should point back to mm_struct
		 * So wee need to find linked list (in generic way)
		 */
//		dbgprint("0x%llX vs 0x%llX",tmp2, tmp4);	
		if (tmp2 == tmp4) {
			int tmp=0;

//			dbgprint("possible vm_area_struct at: 0x%llX",tmp3);

			tmp=simple_list_size(tmp3,image1.pointer_size+(image1.integer_size*2));
			
			if ( tmp > max ) {
				//save local maximum
				printf("	lin_find_process_VMA: vm_area_struct at: 0x%llX, list lize: %d\n",tmp3,tmp);
				max=tmp;
				image1.ts_hl_mm_dst=tmptmp;
				image1.vm_size_dst=image1.pointer_size; 
				image1.vm_list_dst=image1.pointer_size+(image1.integer_size*2); 
			}

		}
	}
//	printf("ts_hl_mm_dst determined as: %d",image1.ts_hl_mm_dst);
	return 0;
}

/* 
 * Function for printing VMA list of process.
 * 
 * list_process_VMA(listhead)
 * listhead is address of list_head of one of the processes
 * 
 * This function should work for any Linux,
 * once determination is completed. 
 *
 * Only printf VMA addresses for now.
 */

int list_process_VMA(u_int64_t addr,FILE * outdump){
	
	u_int64_t tmp1=0,tmp2=0,tmp3=0,tmp4=0;
	int size=0;
	void *buffer=NULL;


//	dbgprint("reading from 0x%llX, offset: %d",addr+image1.ts_hl_mm_dst,image1.ts_hl_mm_dst);

	if (outdump != NULL) { //we will dump memory, so prepare buffer
		if (!(buffer = malloc(MB_1))) { //1MB buffer
			dbgprint("No memory");
			return -2;
		}	
	}	
	if ( (status=image1.read_linear(0,addr+image1.ts_hl_mm_dst,&tmp2,image1.pointer_size,NULL)) != 0 ) { 
		//first read pointer to mm_struct
		errormsg=strerror(errno);
	        dbgprint("read() in %s: %s.\n",inputfile,errormsg);
	}	
//	dbgprint("read 1: 0x%llX",tmp2);

	if ( (status=image1.read_linear(0,tmp2,&tmp1,image1.pointer_size,NULL)) != 0 ) { 
		//read first member of mm_struct structure
		//this should be vm_area_struct *
		errormsg=strerror(errno);
	        dbgprint("read() in %s: %s.\n",inputfile,errormsg);
	}
//	dbgprint("read 2: 0x%llX",tmp1);

	for (;;size++){

//		dbgprint("ptr: 0x%llX, size %d",tmp1,size);

		// get start of area 
		if ( (status=image1.read_linear(0,tmp1+image1.vm_size_dst,&tmp3,image1.integer_size,NULL)) != 0 ) { 
			errormsg=strerror(errno);
	        	dbgprint("read() in %s: %s.\n",inputfile,errormsg);
		}

		// get end of area 
		if ( (status=image1.read_linear(0,tmp1+image1.vm_size_dst+image1.integer_size,&tmp4,image1.integer_size,NULL)) != 0 ) { 
			errormsg=strerror(errno);
		        dbgprint("read() in %s: %s.\n",inputfile,errormsg);
		}
		
		if ((tmp3 == tmp4) && (tmp3 == 0)) {
			printf("		none (kernelspace process?)\n");
			return 0;
		} else {
			// Terrifying 0x%0*llX is for nice padding ;)
			printf("		0x%0*llX-0x%0*llX, (",image1.pointer_size*2,tmp3,image1.pointer_size*2,tmp4); 
			if (outdump != NULL) { //if dumping process memory
				// Dump all areas into one file,
				// ignore distances between areas.
				// TODO coredump format?
				
				u_int64_t tmp5=0,tmp6=0,msize=0;

				// get PGD address first:
				// mm_struct
			        if ((status=image1.read_linear(0,addr+image1.ts_hl_mm_dst,&tmp5,image1.pointer_size,NULL)) != 0 ) {
					errormsg=strerror(errno);
				        dbgprint("read() in %s: %s.\n",inputfile,errormsg);
					return -7; // cannot dump memory
				}
				
				//virtual address of pgd
	                	if ((status=image1.read_linear(0,tmp5+image1.mm_pgd_dst,&tmp6,image1.pointer_size,NULL)) != 0 ) {
					errormsg=strerror(errno);
				        dbgprint("read() in %s: %s.\n",inputfile,errormsg);
					return -8; // cannot dump memory
				}
		                if ((status=image1.read_linear(0,tmp6,NULL,0,&tmp5)) != 0 ) {
					errormsg=strerror(errno);
				        dbgprint("read() in %s: %s.\n",inputfile,errormsg);
					return -9; // cannot dump memory
				}
				// So now we have physical pgd address
//				dbgprint("pgd at: %llx",tmp5);

				msize=tmp4-tmp3;
//				dbgprint("Area size: %llu",msize);

				// We need for, because area can be bigger than buffer
				for (;msize>0;) {
					//read content of memory
					if ( (status=image1.read_linear(tmp5,tmp3,buffer,(msize < MB_1)?msize:MB_1,NULL)) != 0 ) {
						errormsg=strerror(errno);
						dbgprint("read() in %s: %s.\n",inputfile,errormsg);
					}

					// Write content to file
//					dbgprint("writing %llu bytes to output",(msize < MB_1)?msize:MB_1);
					fwrite(buffer,(msize < MB_1)?msize:MB_1,1,outdump);

					if (msize < MB_1) { // unsigned cannot be < 0 ;)
						msize=0;
					} else {
						msize=msize - MB_1; 
					}

				}
			}
		}
		
		//human readable print
		// TODO write macro for B/KB/MB/GB formating
		if (tmp3 > tmp4) { // Negative size, something is wrong..
			printf("!!! Negative size, something is wrong !!! )\n");
		} else
		{
			if ((tmp4-tmp3) > GB_1) {
				printf("%llu GB)\n",(tmp4-tmp3)/(GB_1));
			} else {
				if ((tmp4-tmp3) > MB_1) {
					printf("%llu MB)\n",(tmp4-tmp3)/(MB_1));
				} else 
				{
					if ((tmp4-tmp3) > KB_1) {
						printf("%llu KB)\n",(tmp4-tmp3)/KB_1);
					} else
					{
						printf("%llu Bytes)\n",(tmp4-tmp3));
					}
				}
			}
		}

	
		if (size > MAX_LIST_SIZE ) {
			return -2; //List too long
		} 	

		if ( (status=image1.read_linear(0,tmp1+image1.vm_list_dst,&tmp2,image1.pointer_size,NULL)) != 0 ) {
			errormsg=strerror(errno);
			dbgprint("read() in %s: %s.\n",inputfile,errormsg);
			return -1; //List should be ended by NULL pointer.
		}
//		dbgprint("next list member: %llX",tmp2);	
		if ((tmp1 == tmp2 ) || (tmp2 == 0 )) {
			return (size+1); 
			// It is strange, but sometimes list 
			// seems to be finished by pointer to itself.
		}
		tmp1=tmp2;
	}

	return -5; //Something is wrong
}
/*
 * Function that should list all open files of 
 * 
 * lin_list_files(listhead)
 * return 0 on success 
 * 
 * TODO XXX not working right now
 * Dirty & Dangerous
 */

int lin_list_files(u_int64_t listhead){
	
//	u_int64_t tmp_list,tmp_list2;
//	u_int64_t counter=0,loop_detection=0,loop_detection_heuristic=0;
	u_int64_t tmp1=0,tmp2=0,tmp3=0,tmp4=0,tmp5=0,tmp6=0,tmp7=0; //for listing open files
	char filename[QS_MAGIC_MAX_FILENAME_SIZE]; // tmp filename


// image1.ts_hl_fs_dst=tmp*image1.pointer_size;       //Distance from "tasks" list_head to "files" files_struct in task_struct
// image1.fdt_fd_dst=tmp2*image1.pointer_size;       //Distance from beginning of structure to "fdt" fdtable in files_struct
// image1.fl_fpath_dst=tmp3*image1.pointer_size;       //Distance from beginning of structure to "f_path" path in file structure
// image1.fs_vfs_dst=tmp4*image1.pointer_size;       //Distance from beginning of structure to vfsmount in fs_struct
// image1.one_of_paths=tmpl3;              //Linear pointer to one of path structures

// => task_struct ->  files   ->    fdt   ->  fd    ->  f_path  ->  mnt
//                (files_struct)  (fdtable)  (file **)   (path)   (vfsmount)
//                                                              ->  dentry -> d_name -> name
//                                                                 (dentry)   (qstr)  (char)

	image1.read_linear(0,listhead+image1.ts_hl_fs_dst,&tmp1,image1.pointer_size,NULL); // files_struct

	image1.read_linear(0,tmp1+image1.fdt_fd_dst,&tmp2,image1.pointer_size,NULL); // fdtable struct

	image1.read_linear(0,tmp2+image1.fdt_fd_dst,&tmp3,image1.pointer_size,NULL); // pointer to array of pointers

	//++ for...	
	
	image1.read_linear(0,tmp3/*+i*/,&tmp4,image1.pointer_size,NULL); // one file struct
	
	image1.read_linear(0,tmp4+image1.fl_fpath_dst,&tmp5,image1.pointer_size,NULL); //fpath structure

	image1.read_linear(0,tmp5+image1.pt_dentry_dst,&tmp6,image1.pointer_size,NULL); //dentry structure

	image1.read_linear(0,tmp6+image1.dt_name_dst+image1.qs_name_dst,&tmp7,image1.pointer_size,NULL); // name pointer in qstr structure
	image1.read_linear(0,tmp7,filename,QS_MAGIC_MAX_FILENAME_SIZE,NULL); // name pointer in qstr structure

	// --for
	
	printf("		fd: %llX, name: %s\n",tmp4,filename); //filename length...

	return 0; 
}


/* 
 * Function for printing process list.
 * 
 * list_processes(listhead,distance)
 * listhead is address of list_head of one of the processes
 * distance is in bytes from given listhead structure to comm (char[16]) field
 * 
 * This function should work for any OS with processes in list_head,
 * once determination is completed. 
 *
 * Only printf processes for now.
 */

int list_processes(u_int64_t listhead,int distance){
	
	u_int64_t tmp_list=0,tmp_list2=0, phys_addr=0;
	u_int64_t counter=0,loop_detection=0,loop_detection_heuristic=0;
	char tmp_comm[16];
	int tmp_help=1,first_pattern=1;
	unsigned char *pattern=NULL,*pattern_pos=NULL,*pattern_tmp=NULL; // for creating pattern
//	u_int64_t tmp1,tmp2,tmp3,tmp4,tmp5; //for listing open files

	printf("Listing processes: \n");
/*(linear addresses of kernel, usually ");
	switch(image1.arch) {
		case 1: // i386
			printf("substitute 0xC0000000");
			break;
		case 2: // x86_64
			printf("substitute 0xFFFF880000000000");
			break;
		case 3: // arm
			printf("substitute 0xC0000000");
			break;
	}


	printf("\n to get position in dump file.)\n");
*/
	if (image1.os >= OS_CIRC_HL) {
//		printf("NOTE: Substituting works VERY limited for BSD kernels.\n");
		printf("Disabling VMA listing, not implemented for BSD kernels.\n");
		do_arreas_list=0;
	}


	// Creating process pattern 
	if (do_create_proc_pattern == 1) {
		dbgprint("Creating process pattern");
		if ((pattern=(unsigned char *)malloc(distance)) == NULL) { //pattern itself
			fprintf(stderr, "malloc() failed.\n");
			return -42;
		}
		if ((pattern_tmp=(unsigned char *)malloc(distance)) == NULL) { //pattern to compare
			fprintf(stderr, "malloc() failed.\n");
			return -42;
		}
		if ((pattern_pos=(unsigned char *)malloc(distance)) == NULL) { //list of same bytes
			fprintf(stderr, "malloc() failed.\n");
			return -42;
		}
		memset(pattern_pos,0,distance); // all bytes are ok at beginning
	}



	tmp_list = listhead;
	
	while (tmp_help) { // This is way "down" in list_head (or "up" if bsd-like)	
		image1.read_linear(0,tmp_list,&tmp_list2,image1.pointer_size,&phys_addr);
//		dbgprint("read 0x%llX from 0x%llX",tmp_list2,tmp_list);
		if ((tmp_list == tmp_list2) || (tmp_list2 == 0)) {
			// we point back to ourselves => circle is closed
			// This is fine for bsd
			break;
		}
		else 	
		{
			if (loop_detection_heuristic == 0) {	
				//If we process same task_struct twice. Result not guaranteed.
				loop_detection_heuristic = tmp_list2;
			} else {
				if (loop_detection_heuristic == tmp_list2) {
					break;
				}
			}
			tmp_list = tmp_list2;
			counter++;	//number of processes found
			if (loop_detection++ > MAX_PROC_FOUND) { 
				//there cannot be more than 65K processes on Linux/Windows (ehm;)
				printf("	list_processes: !!! Infinite loop detected !!! Something is wrong\n");
				if (do_create_proc_pattern == 1) {
					free(pattern);
					free(pattern_pos);
					free(pattern_tmp);
				}
				return -3;
			};

//			dbgprint("tmp_list: 0x%llX, tmp_list+distance: 0x%llX",tmp_list,tmp_list+distance);	
			image1.read_linear(0,tmp_list+distance,tmp_comm,16,NULL);
			printf("	process address (virt/phys: 0x%llX/0x%llX) name: %s",tmp_list,phys_addr,tmp_comm); 

			if (do_create_proc_pattern == 1) {
				if (first_pattern == 1) { //save first structure "as is", and diff others
					image1.read_linear(0,tmp_list,pattern,distance,NULL);
					first_pattern=0;
				} else {
					image1.read_linear(0,tmp_list,pattern_tmp,distance,NULL);
					int i;
			/* DBG		printf("\n");
					for (i=0;i<distance;i++){
							printf("%2x ",pattern[i]);
					}
					printf("\n----------\n");
					for (i=0;i<distance;i++){
						if (pattern_pos[i] == 0) {
							printf("%2x ",pattern[i]);
						} else {
							printf(".. ");
						}
					}
					printf("\n");
			*/
					for (i=0;i<distance;i++) {
						// Test for non-zero values
						if ((pattern_pos[i] == 0) && (pattern[i] != pattern_tmp[i]) && (pattern[i]!=0) && (pattern_tmp[i]!=0) ) {
							pattern_pos[i]=2; //  non-zero value
						}	
					
						// Keep only same things
						if ((pattern_pos[i] == 0) && (pattern[i] != pattern_tmp[i])) {
							pattern_pos[i]=1; //  not same for all proc.
						}
						// No non-zero value
						if ((pattern_pos[i] == 2) && ((pattern[i]==0) || (pattern_tmp[i]==0)) ) {
							pattern_pos[i]=1; // non-non-zero;)
						}	

					}
				}
			}
	
			// Default is yes, but user can explicitly disable this
			if (do_arreas_list==1) { 
				printf("\n		mapped areas:\n");
				list_process_VMA(tmp_list,NULL);
			}

			//Now, (if requested) we wil print list of all file descriptors for this process
			if ( do_files_list > 0 ) {
				// This branch is not working right now.  TODO
				//printf(", open files:\n");
				printf("\n");
				//lin_list_files(listhead); //list files for this process
			} else {
				printf("\n");
			}
		}	
	}

	if (do_create_proc_pattern == 1) {
		int i,j;
		printf("----- PROCESS PATTERN START -----\n");

		for (i=0,j=0;i<distance;i++,j=j+2){
//			if ( j > LINE_SIZE) { // To cut long lines
//				j=0;
//				printf("\n");
//			}
			switch(pattern_pos[i]){
				case 0: 
					printf("%02x",pattern[i]);
					break;
				case 1:
					// "bgrep" format
					printf("??"); 
					break;
				case 2: 
					// non-zero value
					printf("NZ");
					break;
			}
		}
		printf("\n----- PROCESS PATTERN END -----\n");
		free(pattern);
		free(pattern_pos);
		free(pattern_tmp);
	}

	printf("Listing processes done.\n");

	return counter;

//	return -5; //Something is wrong
}

/*
 *	We do not search for in init directly, 
 *	because it is too common string in memory => takes too long
 *	This is almost clone of list_processes
 *	
 *	lin_find_userspace_process()
 *	return 0 on success, and set data in global structure
 */

int lin_find_userspace_process(){
	
	u_int64_t tmp_list=0,tmp_list2=0,tmp_addr=0;
	u_int64_t counter=0,loop_detection=0,loop_detection_heuristic=0;
	char tmp_comm[16];
	int tmp_help=1;

	printf("	lin_find_userspace_process: Looking for userspace process \"%s\".\n",MAGIC_USER_PROCESS); 
	
	tmp_list = image1.ts_one_of_lists; //This is already from list_head, but not userspace

	while (tmp_help) { //This is way "down" in list_head
	
		image1.read_linear(0,tmp_list,&tmp_list2,image1.pointer_size,NULL);
//		printf("	lin_find_userspace_process: tmp:%llX, tmp2:%llX.\n",tmp_list,tmp_list2); 

		if ((tmp_list == tmp_list2) || (tmp_list2 == 0)) { //List is empty, or invalid linear address was provided 
			printf("	lin_find_userspace_process: strange,list1: %llX,list2: %llX\n",tmp_list,tmp_list2);
			printf("	lin_find_userspace_process: => calling debug().\n");
			debug();
			tmp_help=0;
		} else // we point back to ourselves => circle is closed	
		{
			if (loop_detection_heuristic == 0) {	//If we process same task_struct twice. No guaranteed result
				loop_detection_heuristic = tmp_list2;
			} else {
				if (loop_detection_heuristic == tmp_list2) {
					printf("	lin_find_userspace_process: Not found \"%s\", something is wrong.\n",MAGIC_USER_PROCESS);
					return -1;
				}
			}
			tmp_list = tmp_list2;
			counter++;	//number of processes found
			if (loop_detection++ > MAX_PROC_FOUND) { 
				//there cannot be more than 65K processes on Linux (ehm;)
				printf("	lin_find_userspace_process: !!! Infinite loop detected !!! Something is wrong\n");
				return -3;
			};
	
			image1.read_linear(0,tmp_list+image1.ts_hl_comm_dst,tmp_comm,16,&tmp_addr);
			//printf("Found userspace process \"%s\" at 0x%llX\n",tmp_comm,tmp_list); //DBG

			if ( (strncmp(tmp_comm, MAGIC_USER_PROCESS, strlen(MAGIC_USER_PROCESS))) == 0 ) {
				printf("Found userspace process \"%s\" at (virt/phys) 0x%llX/0x%llX\n",MAGIC_USER_PROCESS,tmp_list,tmp_addr); 
				image1.ts_one_of_lists=tmp_list;
				return 0;
			}
		}	
	}

	return -5; //Something is wrong, list is empty or invalid
}

/*  
 *  Function for getting size of circular list head .
 *  ( used in linux/win process structures )
 *
 *  Whole list is traversed, until first member is seen 
 *  second time. Or until loop detection is successful.
 * 
 *  input: Physical address of list_head
 *  output: Number of members, negative value if problem is detected
 */

int lin_list_head_size(u_int64_t listhead){
	
	u_int64_t tmp_list=0,tmp_list2=0;
	u_int64_t tmp_list3=0,tmp_list4=0; //backward checking
	u_int64_t counter=0,errcounter=0,loop_detection=0,loop_detection_heuristic=0;
	int tmp_help=1;

	if ( (status=lseek(inputfd,listhead,SEEK_SET)) == -1 ) {  
		errormsg = strerror(errno);
	        fprintf(stderr, "lin_list_head_size: lseek() to 0x%llX failed because of %s.\n",listhead,errormsg);
	        return(-1);
	}
	
	if ( (status=read(inputfd,&tmp_list,image1.pointer_size)) == -1 ) {
	        errormsg=strerror(errno);
	        fprintf(stderr, "lin_list_head_size: read() in %s: %s.\n",inputfile,errormsg);
	        return(-2);
	}
//	dbgprint(" 0x%llX\n",listhead);

	while (tmp_help) { //This is way "down" in list_head
		image1.read_linear(0,tmp_list,&tmp_list2,image1.pointer_size,NULL);//forward pointer
		image1.read_linear(0,tmp_list+image1.pointer_size,&tmp_list3,image1.pointer_size,NULL); //backward pointer

		// if we have already seen this process (circle closed)
		if ((tmp_list == tmp_list2) || (tmp_list2 == 0)) {tmp_help=0;} else 
		{
			//If we process same task_struct twice. No guaranteed result
			if (loop_detection_heuristic == 0) {	

				loop_detection_heuristic = tmp_list2;
			} else {
				if (loop_detection_heuristic == tmp_list2) { 
					//inner circle closed
//					dbgprint("counter:%d.",counter);
					return counter;
				}
			}
			// Check if we are really in double linked list, or just jumping 
			// over some addresses
			if ((tmp_list3 != tmp_list4) && (tmp_list4 != 0)) {
				//tolerate some number of errors
				errcounter++;
//				dbgprint("errcounter ++ (%d), tmp3: %llX,tmp4: %llX, tmp2: %llX,tmp: %llX",errcounter, tmp_list3, tmp_list4, tmp_list2, tmp_list);	

				if (errcounter > TS_MAGIC_LIST_MAX_ERR){
//					dbgprint("errcounter reached (%d)",errcounter);	
					return -7;
				}
			}

			tmp_list4 = tmp_list; //back pointer checking
			tmp_list = tmp_list2;
			counter++;	//number of processes found
			if (loop_detection++ > MAX_PROC_FOUND) { //there cannot be more than 65K processes on Linux (ehm;)
				printf("	lin_list_head_size: !!! Infinite loop detected !!! Something is wrong\n");
				return -3;
			};
	
		}	
	}

	return -5; //Something is wrong
}

/*  
 *  Function for getting size of bsd-like list.
 *  ( used in BSD "proc" structure )
 *
 *                   ||
 *                   ||
 *                   \/ 
 *  INIT |--------> NEXT |-------->  |-------->  NULL
 *  NULL <--------| PREV <--------|              ...
 *
 *
 * from sys/queue.h
 * ...
 * #define LIST_HEAD(name, type)
 * struct name {
 *         struct type *lh_first;  // first element
 * }
 * 
 * #define LIST_HEAD_INITIALIZER(head)
 *         { NULL }
 * 
 * #define LIST_ENTRY(type)
 * struct {
 *         struct type *le_next;   // next element
 *         struct type **le_prev;  // address of previous next element
 * }
 * ...
 *
 *  bsd_list_head_size (address to verify , output parameter "first member of list" )
 *  return number of members on success, otherwise error < 0
 *  if list_head_start is not NULL, pointer to first member of 
 *  list is stored here.
 * 
 *
 *  Because this list is not cycled, first we must travel to one side
 *  until NULL is encountered, and then travel to the other side,
 *  looking for NULL pointer. We also need to propagate information 
 *  where list starts
 *
 *  input: Physical address of list_head, offset to string (optional)
 *  output: Number of members ( <0 if error occurs ), start of list (optional), 
 *          total size of all strings at offset (optional)
 * 
 */

int bsd_list_head_size(u_int64_t listhead, u_int64_t * list_head_start, int offset, int * strings_size){
	
	u_int64_t tmp_list=0,tmp_list2=0;
	u_int64_t tmp_list3=0,tmp_list4=0; //backward checking
	u_int64_t tmp_list5=0,tmp_list6=0;
	u_int64_t loop_detection_heuristic=0;
	int counter=0,errcounter=0;
	int tmp_help=1;
	char tmp_comm[TS_COMM_LEN]; //structure name (string)

//	dbgprint("listhead: 0x%llX, offset %d, (list_head_start: 0x%llX, strings_size: %d) ",listhead,offset,*list_head_start,*strings_size);

	if ( (status=lseek(inputfd,listhead,SEEK_SET)) == -1 ) {  
		errormsg = strerror(errno);
	        fprintf(stderr, "lin_list_head_size: lseek() to 0x%llX failed because of %s.\n",listhead,errormsg);
	        return(-1);
	}
	
	if ( (status=read(inputfd,&tmp_list,image1.pointer_size)) == -1 ) {
	        errormsg=strerror(errno);
	        fprintf(stderr, "lin_list_head_size: read() in %s: %s.\n",inputfile,errormsg);
	        return(-2);
	}
//	dbgprint(" 0x%llX, tmp_list: 0x%llX",listhead,tmp_list);

	// counting length of all list_members strings. 
	// See explanation before this function.
	if ((offset != 0) && (strings_size != NULL)) {
		*strings_size=0;
	}

	loop_detection_heuristic = tmp_list; //loop detection
	// first up and first down back-pointer checks will fail
	// ok, just ignore them


	while (tmp_help) { //This is way "up" in list_head

		// forward pointer
		image1.read_linear(0,tmp_list,&tmp_list2,image1.pointer_size,&tmp_list5);
		// backward pointer
		image1.read_linear(0,tmp_list+image1.pointer_size,&tmp_list3,image1.pointer_size,&tmp_list6);
//		dbgprint("up-step, 0x%llX (0x%llX), 0x%llX (0x%llX)", tmp_list2,tmp_list5,tmp_list3,tmp_list6);

		// Counting length of all list_members strings. 
		// See explanation before this function.
		if (offset != 0) {
			// entry name
			image1.read_linear(0,tmp_list+offset,tmp_comm,TS_COMM_LEN,&tmp_list5);
			*strings_size+=strnlen(tmp_comm,TS_COMM_LEN);
//			dbgprint("List 0x%llX at 0x%llx (0x%llX) proc: %s",tmp_list,tmp_list+offset,tmp_list5,tmp_comm); 
		}

		if ((tmp_list == tmp_list2) || (tmp_list2 == 0)) {
			tmp_help=0; // we have already seen this process or NULL
		} else {
			if (loop_detection_heuristic == tmp_list2) { //circle closed
				printf("	bsd_list_head_size: Suspicious structure at 0x%llX, loop detected. Size :%d.\n",tmp_list,counter);
				return -9;
			}
			
			// Check if we are really in double linked list, or just jumping 
			// over some addresses
			if ((tmp_list3 != tmp_list4) && (tmp_list4 != 0)) {
				//tolerate some number of errors
				errcounter++;
				dbgprint("errorcounter ++ (%d), tmp3: %llX,tmp4: %llX, tmp2: %llX,tmp: %llX",errcounter, tmp_list3, tmp_list4, tmp_list2, tmp_list);	
				if (errcounter > TS_MAGIC_LIST_MAX_ERR){
					dbgprint("errorcounter reached (%d)",errcounter);	
					return -7;
				}
			}

			tmp_list4 = tmp_list; //back pointer checking
			tmp_list = tmp_list2;
			counter++;	//number of processes found
			if (counter > MAX_PROC_FOUND) { //there cannot be more than 65K processes on BSD (ehm;)
				printf("	bsd_list_head_size: !!! Infinite loop detected !!! Something is wrong\n");
				return -3;
			};
	
		}	
	}

	// Do same thing for way "down"
//	dbgprint("down: 0x%llX, tmp_list: 0x%llX\n",listhead,tmp_list);
	tmp_list = loop_detection_heuristic;	//original pointer
	tmp_help=1;

	while (tmp_help) { //This is way "down" in list_head
		image1.read_linear(0,tmp_list,&tmp_list3,image1.pointer_size,&tmp_list6);//forward pointer
		image1.read_linear(0,tmp_list+image1.pointer_size,&tmp_list2,image1.pointer_size,&tmp_list5);//backward pointer
//		dbgprint("down-step, 0x%llX (0x%llX), 0x%llX (0x%llX)", tmp_list2,tmp_list5,tmp_list3,tmp_list6);

		// counting length of all list_members strings. 
		// See explanation before this function.
		if (offset != 0) {
			// entry name 
			image1.read_linear(0,tmp_list+offset,tmp_comm,TS_COMM_LEN,&tmp_list5);
			*strings_size+=strnlen(tmp_comm,TS_COMM_LEN);
//			dbgprint("List 0x%llX at 0x%llx (0x%llX) proc: %s",tmp_list,tmp_list+offset,tmp_list5,tmp_comm); 
		}

		if ((tmp_list == tmp_list2) || (tmp_list2 == 0)) {
			tmp_help=0; // we have already seen this process or NULL
		} else {
			if (loop_detection_heuristic == tmp_list2) { //circle closed
				return -9;
			}
			
			// Check if we are really in double linked list, or just jumping 
			// over some addresses
			if (tmp_list3 != tmp_list4) {
				//tolerate some number of errors
				errcounter++;
//				dbgprint("errorcounter ++ (%d), tmp3: %llX,tmp4: %llX, tmp2: %llX,tmp: %llX",errcounter, tmp_list3, tmp_list4, tmp_list2, tmp_list);	
				if (errcounter > TS_MAGIC_LIST_MAX_ERR){
//					dbgprint("errorcounter reached (%d)",errcounter);	
					return -7;
				}
			}

			tmp_list4 = tmp_list; //back pointer checking
			tmp_list = tmp_list2;
			counter++;	//number of processes found
			if (counter > MAX_PROC_FOUND) { //there cannot be more than 65K processes on BSD (ehm;)
				printf("	bsd_list_head_size: !!! Infinite loop detected !!! Something is wrong\n");
				return -3;
			};
	
		}	
	}
	
	// return first member of list
	if (list_head_start != NULL) {
		* list_head_start=tmp_list;
	}
	return counter; // whole size
}



/*
 * Function for printing list of modules.
 *
 * lin_list_modules(listhead, distance)
 * listhead is address of list_head of one of the modules (not linear address, but address in physical dump file)
 * distance is in bytes from given listhead structure to comm (char[16]) field.
 *
 * Only printf modules for now.
 */


int lin_list_modules(u_int64_t listhead,int distance){
	
	u_int64_t tmp_list=0,tmp_list2=0,tmp_addr=0;
	u_int64_t counter=0,loop_detection=0,loop_detection_heuristic=0;
	char tmp_comm[MD_NAME_LEN]; //according to module.h, with  reserve
	int tmp_help=1,first_pattern=1;
	unsigned char *pattern=NULL,*pattern_pos=NULL,*pattern_tmp=NULL; // for creating pattern

	// Creating module pattern 
	if (do_create_module_pattern == 1) {
		dbgprint("Creating module pattern");
		if ((pattern=(unsigned char *)malloc(distance+MD_PATTERN_SPREAD)) == NULL) { //pattern itself
			fprintf(stderr, "malloc() failed.\n");
			return -42;
		}
		if ((pattern_tmp=(unsigned char *)malloc(distance+MD_PATTERN_SPREAD)) == NULL) { //pattern to compare
			fprintf(stderr, "malloc() failed.\n");
			return -42;
		}
		if ((pattern_pos=(unsigned char *)malloc(distance+MD_PATTERN_SPREAD)) == NULL) { //list of same bytes
			fprintf(stderr, "malloc() failed.\n");
			return -42;
		}
		memset(pattern_pos,0,distance+MD_PATTERN_SPREAD); // all bytes are ok at beginning
	}


	printf("Listing modules:\n");
	tmp_list = listhead;

	while (tmp_help) { //This is way "down" in list_head
	
		image1.read_linear(0,tmp_list,&tmp_list2,image1.pointer_size,NULL);


		if (do_create_module_pattern == 1) {
			if (first_pattern == 1) { //save first structure "as is", and diff others
				image1.read_linear(0,tmp_list,pattern,distance+MD_PATTERN_SPREAD,NULL);
				first_pattern=0;
			} else {
				image1.read_linear(0,tmp_list,pattern_tmp,distance+MD_PATTERN_SPREAD,NULL);
				int i;
				for (i=0;i<distance+MD_PATTERN_SPREAD;i++) { 
					// Test for non-zero values
					if ((pattern_pos[i] == 0) && (pattern[i] != pattern_tmp[i]) && (pattern[i]!=0) && (pattern_tmp[i]!=0) ) {
						pattern_pos[i]=2; //  non-zero value
					}	
					
					// Keep only same things
					if ((pattern_pos[i] == 0) && (pattern[i] != pattern_tmp[i])) {
						pattern_pos[i]=1; //  not same for all proc.
					}
					// No non-zero value
					if ((pattern_pos[i] == 2) && ((pattern[i]==0) || (pattern_tmp[i]==0)) ) {
						pattern_pos[i]=1; // non-non-zero;)
					}	

				}
			}
		}



		if ((tmp_list == tmp_list2) || (tmp_list2 == 0)) {
			tmp_help=0;
		} else {// we point back to ourselves => circle has closed	

			//If we process same module twice. No guaranteed result
			if (loop_detection_heuristic == 0) {
				loop_detection_heuristic = tmp_list2;
			} else {
				if (loop_detection_heuristic == tmp_list2) {
					if (do_create_module_pattern == 1) {
						int i;
						printf("----- MODULE PATTERN START -----\n");
		
						for (i=0;i<distance+MD_PATTERN_SPREAD;i++){

						// I decided leave patterns in one line,
						// so they can be easily cut&pasted
						//	if ( j > LINE_SIZE) { // To cut long lines
						//		j=0;
						//		printf("\n");
						//	}
							switch(pattern_pos[i]){
								case 0: 
									// hex value
									printf("%02x",pattern[i]);
									break;
								case 1:
									// "bgrep" format
									printf("??"); 
									break;
								case 2: 
									// non-zero value
									printf("NZ");
									break;
							}
						}
						printf("\n----- MODULE PATTERN END -----\n");
						free(pattern);
						free(pattern_pos);
						free(pattern_tmp);
					}
		
					printf("Listing modules done.\n");
					return counter;
				}
			}
			tmp_list = tmp_list2;
			counter++;	// number of modules found
			if (loop_detection++ > MAX_PROC_FOUND) { 
				// there cannot be more than 65K modules on Linux (ehm;)
				printf("	list_modules: !!! Infinite loop detected !!! Something is wrong\n");
				return -3;
			};
	
			image1.read_linear(0,tmp_list+distance,tmp_comm,MD_NAME_LEN,&tmp_addr);
			printf("	module address (virt/phys): 0x%llX/0x%llX, name: %s\n",tmp_list,tmp_addr,tmp_comm); 
		}	
	}

	return -5; // Something is wrong
}

/*
 * Function that tries to determine processes specific information 
 * about analyzed image.
 *
 * Some process (and if analyzing linux then some userland process) 
 * are found, and distance between important fields are determined.
 *
 * If successful enough, OS is determined as side effect.
 *
 * Information about "module" structure are determined separately,
 * because module support is only optional, and on some systems
 * there are almost no modules.
 *
 * proc_determine ()
 * return 0 on success, otherwise error <> 0
 * Discovered information are saved in global structure image1.
 *
 * */

int proc_determine(){
	
	int tmp;
	u_int64_t tmp2=0;
	
	printf("	proc_determine: Looking for process \"%s\".\n",MAGIC_PROCESS);
        
 	if (search_init(inputfile,MAGIC_PROCESS,0) != 0 ) { // This is not optimal
		printf("proc_determine: search_init returned error. Aborting.\n");
		return -1;
	}

	while ((tmp2=search_next())!=0){
//		dbgprint("Trying to verify: 0x%llX as task_struct",tmp2);
		if (verify_task_struct(tmp2) == 0) { //we have found valid process
			printf("Found valid process at 0x%llX.\n",tmp2);
			break;
		}
	}

	if (tmp2 == 0) { //Found nothing
		if (strncmp(LINUX_MAGIC_PROCESS,MAGIC_PROCESS,strlen(LINUX_MAGIC_PROCESS)) == 0 ) {
			// if user did not used --magic-process option
			// Try all again with windows process
			// This takes twice long, but we have "automatic" detection of OS;) 
			fprintf(stderr,"	proc_determine: \"%s\" not found, second try: \"%s\"\n",MAGIC_PROCESS,WIN_MAGIC_PROCESS);
		
			search_exit(); //close file
		 	if (search_init(inputfile,WIN_MAGIC_PROCESS,0) != 0 ) { //This seems to work, even if it is not optimal
				printf("proc_determine: search_init returned error. Aborting.\n");
				return -2;
			}

			while ((tmp2=search_next())!=0){
//				printf("	proc_determine: Trying to verify: %llX as task_struct\n",tmp2);
				if (verify_task_struct(tmp2) == 0) { //we have found valid process
					// NOTE: verify_task_struct works even for other OS.
					// naming of function will be corrected in next release.
					printf("Found valid windows process at 0x%llX.\n",tmp2);
					strncpy(image1.operating_system,"windows\0",8);
					break;
				}
			}
		}
		if (tmp2 == 0) { //Found nothing (again)
			printf("	proc_determine: Cannot find process, exiting.\n"); 
			return 2;
		}
	}else{
		// We found MAGIC_PROCESS. 
		// If user did not specified magic_process or OS, then we now it is linux.
		if (strncmp(image1.operating_system,"unknown",7) == 0){
			if (strncmp(MAGIC_PROCESS,LINUX_MAGIC_PROCESS,strlen(LINUX_MAGIC_PROCESS)) == 0){	
				strncpy(image1.operating_system,"linux\0",6);	
			}
		}
	}


//	dbgprint("image1.os: %d",image1.os);
	if ((image1.os <= OS_CIRC_HL) && (image1.os != OS_WINDOWS)) {  
		// If it is not Windows.. (but does not use BSD style list_head)
		// (on win no userspace processes are required)
			
		// Find some userspace process
		if ( (tmp=lin_find_userspace_process()) !=0) {
			printf("	linux_determine: lin_find_userspace_process returned: %d, assuming windows OS.\n",tmp);
			strncpy(image1.operating_system,"windows\0",8); 
			// no init process => probably win? 
			// this does not have to be true, 
			// but don't waste our results and treat it as win.
			// return 3;
			printf("Determination process of task_struct complete.\n");
			return 0;
		}
	
		//determine distances of VMAs
		if ( (tmp=lin_find_process_VMA(image1.ts_one_of_lists)) !=0) {
			printf("	linux_determine: lin_find_process_VMA returned: %d, terminating.\n",tmp);
			return 4;
		}


		// VMAa are determined before process VPT,
		// because finding mm_struct offset inside task_struct is more reliable
		

		// determine VPT distances
		if ( (tmp=lin_find_process_VPT(image1.ts_one_of_lists)) !=0) {
			printf("	linux_determine: lin_find_process_VPT returned: %d.\n",tmp);
			printf("	linux_determine: Reading of userspace addresses will be not available..\n");
		}
		


     	} 

	printf("Determination process of task_struct complete.\n");
	return 0;
}


/*
 * Function that tries to determine linux module-specific 
 * information about analyzed image.
 *
 * Some module is found,
 * and distance between important fields are determined.
 *
 * This is done separately,
 * because module support is only optional, and is not 
 * critical for another analyze.
 *
 * linux_determine_module ()
 * return 0 on success, otherwise error <> 0
 * Discovered information are saved in global structure image1.
 *
 * */

int lin_linux_determine_module(){
	
	u_int64_t tmp2=0;
	
	printf("	linux_determine_module: Looking for module \"%s\".\n",MAGIC_MODULE);
	printf("	linux_determine_module: Starting at address 0x%X.\n",MODULE_MAGIC_OPT_ADDR);

 	if (search_init(inputfile,MAGIC_MODULE,MODULE_MAGIC_OPT_ADDR) != 0 ) { 
		printf("linux_determine: search_init returned error. Aborting.\n");
		return -1;
	}

	
	while ((tmp2=search_next())!=0){
//		printf("	linux_determine: Trying to verify: %llX as task_struct\n",tmp2);
		if (lin_verify_module(tmp2) == 0) { //we have found valid module

			printf("Found valid module at 0x%llX.\n",tmp2);
			printf("Determination process of module structure complete.\n");
			return 0;
		}
	}

	//Found nothing
	printf("	linux_determine_module: Cannot find module %s while using optimization.\n",MAGIC_MODULE); 
	printf("	linux_determine_module: Retrying with whole file (slower).\n");

	if (search_init(inputfile,MAGIC_MODULE,0) != 0 ) { 
		//Search whole file
		printf("linux_determine: search_init returned error. Aborting.\n");
		return -1;
	}
	
	while ((tmp2=search_next())!=0){
		if (lin_verify_module(tmp2) == 0) { //we have found valid module
			printf("Found valid module at 0x%llX.\n",tmp2);
			printf("Determination process of module structure complete.\n");
			return 0;
		}
	}

	//Found nothing
	printf("	linux_determine_module: Cannot find module %s.\n",MAGIC_MODULE); 
	return 2;
}

/*
 * Function that prints detailed information 
 * about certain process.
 *
 * print_process_info ( process address, filename to store process memory)
 * return 0 on success, otherwise error <> 0
 *
 * */

int print_process_info(u_int64_t single_process, char * dump_filename)
{
	FILE *outdump;
	int status900=0,i=0;	
	u_int64_t tmp901=0,tmp902=0,tmp903=0,tmp904=0;
	u_int64_t tmp905=0,tmp906=0,tmp907=0,tmp908=0; //pointer to previous and next process
	unsigned char tmp900_comm[TS_COMM_LEN];

	// physical address of list_head
        status900=image1.read_linear(0,single_process,NULL,0,&tmp901);
	// process name
        status900=image1.read_linear(0,single_process+image1.ts_hl_comm_dst,tmp900_comm,TS_COMM_LEN,NULL);
	//mm_struct
        status900=image1.read_linear(0,single_process+image1.ts_hl_mm_dst,&tmp902,image1.pointer_size,NULL);
//	dbgprint("mmstruct at %llX, ts_hl_mm_dst: %d, mm_pgd_dst: %d",tmp902,image1.ts_hl_mm_dst,image1.mm_pgd_dst);
	if (status900 == 0 ) {
		//pgd
		image1.read_linear(0,tmp902+image1.mm_pgd_dst,&tmp903,image1.pointer_size,&tmp904);
		//physical address of pgd
		status900=image1.read_linear(0,tmp903,NULL,0,&tmp904);
	} 

	//ptr to previous proc
	image1.read_linear(0,single_process,&tmp905,image1.pointer_size,&tmp906); 
	//ptr to next proc
	image1.read_linear(0,single_process+image1.pointer_size,&tmp907,image1.pointer_size,&tmp908);

	printf("Detailed information about process at 0x%llX.\n",single_process);
	printf("Process linear address(in kernel context): 0x%llX\n",single_process); 
	printf("Process physical address(offset in dump): 0x%llX\n",tmp901);
	printf("Process name(short): \"%s\"\n",tmp900_comm);
	printf("Process PGD address(linear): 0x%llX\n",tmp903);
	printf("Process PGD address(physical): 0x%llX\n",tmp904);
	printf("Previous process: 0x%llX (phys: 0x%llX)\n",tmp905,tmp906);
	printf("Next process: 0x%llX (phys: 0x%llX)\n",tmp907,tmp908);
	printf("Process mapped areas (userspace addresses):\n");
	if (dump_filename == NULL) {
		list_process_VMA(single_process,NULL);
	} else {	
		printf("Saving process memory into %s:\n",dump_filename);

		if ( (outdump = fopen(dump_filename,"w")) == NULL ) {
			errormsg = strerror(errno);
			fprintf(stderr, "%s: %s.\n",statusfile,errormsg);
			return 22;
		}

		list_process_VMA(single_process,outdump);

		i=fclose(outdump);

		printf("Process memory saved into %s:\n",dump_filename);

	}
	printf("Process open files: TODO;)\n");
//      i386_lin_list_files(listhead); //list files for this process	//TODO
	return i;

}

/*
 * Function that prints detailed information 
 * about certain module.
 *
 * print_module_info ()
 * return 0 on success, otherwise error <> 0
 *
 * */

int print_module_info(u_int64_t single_module)
{
	u_int64_t tmp801=0;		
	unsigned char tmp800_comm[MD_NAME_LEN];

	image1.read_linear(0,single_module+image1.md_hl_name_dst,tmp800_comm,MD_NAME_LEN,&tmp801); 
	printf("Detailed information about module at 0x%llX.\n",single_module);
	printf("Module linear address(in kernel context): 0x%llX\n",single_module); 
	printf("Module physical address(offset in dump): 0x%llX\n",tmp801);
	printf("Module name(short): \"");
	int tmp802=0;
	for(;tmp802<=MD_NAME_LEN;tmp802++) {
		int tmp803=tmp800_comm[tmp802];
		if (isprint(tmp803) && ( tmp803 == ' ' || !isspace(tmp803)) ) {
		//only printable chars, but not end of lines, but yes space;)
			putc(tmp803,stdout);
		} else {
			putc('.',stdout); //non printable (in ascii)
		}
	}
	printf("\"\n");	
	return 0;	
}
