/*
 *	Functions for analysing i386 architecture systems
*/

#include "../../globals.h"
#include "magic.h"

/*
 * Some alchemy:
 */

/*
 * Function that manages mapping from linear to physical address space.
 * 
 * priv_block_size from lin_addr are read using provided PGD (kernel PGD by default) into out_data pointer.
 * dump_offset is output parameter, if NULL, ignored.
 * Otherwise physical offset translated from linear address is stored here.
 * (!) Please note, that read block can be larger than one page. 
 * In this case offset is valid only for first part of data.
 */

int i386_read_linear(u_int64_t PGD, u_int64_t lin_addr, void *out_data, u_int64_t priv_block_size, u_int64_t *dump_offset){

	//"PGD" is needed, because each process have it's own linear address space. Process 0 means kernel memory.

	//Find in VPT "physical" address, where data should be. (if problem, return 1)
	//Read block_size of that data into out_data (if problem, return 2)
	//return 0

	//u_int64_t PDE_addr; //= find_PDE();

	u_int64_t physical_offset = lin_addr;
	u_int64_t PDE_offset = lin_addr;
	u_int64_t PTE_offset = lin_addr;
	u_int64_t page_number = lin_addr;
	u_int64_t tmp_var = 0;

//	dbgprint("i386_read_linear sizeoff(\n PGD 0x%u\nlin_addr 0x%u\nout_data 0x%u\npriv_block_size %u\ndump_offset 0x%u)\n",sizeof(PGD),sizeof(lin_addr),sizeof(out_data),sizeof(priv_block_size),sizeof(dump_offset)); 

//	 dbgprint("i386_read_linear(\nPGD 0x%llX\nlin_addr 0x%llX\nout_data %p\npriv_block_size %llu\ndump_offset %p)\n",PGD,lin_addr,out_data,priv_block_size,dump_offset); 

	if (PGD == 0) {	//special case
		PGD = image1.VPT_address;

	/* if VPT_address is not available,
	 * we can try "best effort", and use hard-coded 
	 * "direct mapping", usually used by OS kernels.
	 * (and OS specific)
	 * for linux on i386, 0xC0000000 can be subtracted 
	 * from linear kernelspace addresses.
	 * Does not work for BSD (!)
  	 */  
		if (image1.VPT_not_available == 1) { 
			//kernel text direct mapping
			if ((lin_addr > 0xC0000000) && 
			    (lin_addr < 0xFFFFFFFF)) {
				physical_offset=lin_addr - 0xC0000000;
				dbgprint("direct mapping 0x%llX -> 0x%llX",lin_addr,physical_offset);
			}
			
			// If none of mappings matches 
			if (physical_offset == lin_addr) {
//				dbgprint("nothing matched, 0x%llX",lin_addr);
				return -1;
			}
			
			// Read memory
			lseek(inputfd,physical_offset,SEEK_SET); //level 0 (physical page) 
//			dbgprint("(level 0)Reading from PO: 0x%llX\n",physical_offset);
			
			if (dump_offset != NULL) { //save physical offset to output variable
				*dump_offset=physical_offset; 
			}

			if ( (status=read(inputfd,out_data,priv_block_size)) == -1 ) {
        		        errormsg=strerror(errno);
               			fprintf(stderr, "read_liner: read() in %s: %s.\n",inputfile,errormsg);
	                	return 38;
			}

			return 0;
		}

	} // else PGD is linear address of PGD of some process


//	dbgprint("PGD1: %llX, addr: %llX",PGD,lin_addr);
	
	//Kernel linear address space:

	// For extracting addresses I use bitwise and, for extracting offsets I have to shift bits.
        physical_offset = physical_offset & 0x003FFFFF;//  << 10; >>10;//bits 21:0  // 4-Mbyte Page Translation
        PDE_offset = PDE_offset >> 22; //bits 31:22 

//	dbgprint("lin_addr: %llX, Phys_Off: 0x%llX, PDE_Off: 0x%llX, PDE_e: 0x%llX\n",lin_addr,physical_offset,PDE_offset,PGD+PDE_offset*image1.pointer_size); 

        lseek(inputfd,PGD+PDE_offset*image1.pointer_size,SEEK_SET);// *4 because each entry have 32 bits = 4bytes
		
        if ( (status=read(inputfd,&page_number,image1.pointer_size)) == -1 ) {
                errormsg=strerror(errno);
                fprintf(stderr, "read_liner: read() in %s: %s.\n",inputfile,errormsg);
                return 35;
        }

//	dbgprint("page_number: 0x%X, pointer size: %d\n",page_number,image1.pointer_size);

	// TODO: process all necessary special bit, like swapped, etc...

	tmp_var = page_number;
	tmp_var = tmp_var & 0x00000080; //only bit NUMBER 7 (ie. 8th bit)

	if (tmp_var != 0) { //4MB pages are used
		
//		dbgprint("4MB pages used for this page.\n"); 

		page_number = page_number >> 22; //bits 31:22

		//If we want to read more than 1 page:
		// (use recursion;)
		//
		if (priv_block_size > (MB_4-physical_offset)){ // offset + block_size must be less then 4M 
//			printf("i386_lin_read_linear: page too small for reading... processing another one.\n");
			//recursion
			
			tmp_var = MB_4-physical_offset;

			i386_read_linear(PGD,lin_addr+tmp_var,out_data+tmp_var,priv_block_size-tmp_var,NULL);

		} else{
			tmp_var=priv_block_size;
		}	

		lseek(inputfd,page_number*MB_4+physical_offset,SEEK_SET); //(4MB page number "large_page_number" + offset)
//		dbgprint("4MB page_number: 0x%llX, phys. addr: 0x%llX\n",page_number,page_number*MB_4+physical_offset);

		if (dump_offset != NULL) { //save physical offset to output variable
			*dump_offset=page_number*MB_4+physical_offset;
		}
		
		if ( (status=read(inputfd,out_data,tmp_var)) == -1 ) {
        	        errormsg=strerror(errno);
                	fprintf(stderr, "i386_read_linear: read() in %s: %s.\n",inputfile,errormsg);
	                return 35;
        	}
	} else {   //4KB pages are used
//		dbgprint("4KB pages used for this page.\n");
	
		page_number = page_number & 0xFFFFF000; //bits 31:12 ... This must be done this way, not like commented out example

		PTE_offset = (lin_addr & 0x003FF000) >> 12; //bits 21:12 ... This is offset, so we have to shift bits.// optimize?
		
		physical_offset = physical_offset & 0x00000FFF;//  << 10; >>10;//bits 11:0
		
//		dbgprint("4KB Searching for PTE in 0x%llx.\n",page_number);

		lseek(inputfd,page_number+PTE_offset*image1.pointer_size,SEEK_SET); 

		if ( (status=read(inputfd,&tmp_var,image1.pointer_size)) == -1 ) {
                 	errormsg=strerror(errno);
		        fprintf(stderr, "read_liner: read() in %s: %s.\n",inputfile,errormsg);
			return 36;
		}
		
//		dbgprint("4KB In PTE I found 0x%llX.\n",tmp_var);

		//tmp_var = tmp_var >> 12; //bits 31:12  //PTE entry
		tmp_var = tmp_var & 0xFFFFF000; // this or this way?

//		dbgprint("4KB Seeking to address 0x%llX.\n",tmp_var+physical_offset);

		lseek(inputfd,tmp_var+physical_offset,SEEK_SET);

		if (dump_offset != NULL) { //save physical offset to output variable
			*dump_offset=tmp_var+physical_offset;
		}

		//If we want to read more than 1 page:
		// (use recursion;)
		//
		if (priv_block_size > (KB_4-physical_offset)){ // offset + block_size must be less then 4M
//			printf("i386_lin_read_linear: page too small for reading... processing another one.\n");
			//recursion
			
			tmp_var = KB_4-physical_offset; //

			//Process rest of request recursively... beware about effectiveness
			i386_read_linear(PGD,lin_addr+tmp_var,out_data+tmp_var,priv_block_size-tmp_var,NULL);

		} else{
			tmp_var=priv_block_size;
		}

		if ( (status=read(inputfd,out_data,tmp_var)) == -1 ) {
        	        errormsg=strerror(errno);
                	fprintf(stderr, "read_liner: read() in %s: %s.\n",inputfile,errormsg);
	                return 38;
        	}


	}

	return 0;

}

#if 0

TODO
Possible on i386?

/*
 * Function for verifying if given address contains
 * PDPE from page tables.
 *
 * Verification is done by simple checking
 * if all entries contain only zeroes or 
 * if they contain entry with certain valid bits.
 * This should usually be true, but it is just heuristic, 
 * and can fail.
 *
 * x86_64_looks_like_page_table (physical address to verify )
 * return 0 on success, otherwise error > 0
 *
 */


int i386_looks_like_page_table(u_int64_t my_addr){

	int mi=0,ecounter=0,rcounter=0; //error counter
	u_int64_t tmp_small_page[I386_VPT_MAGIC_BS/I386_POINTER_SIZE]; //4KB

//	dbgprint("trying  0x%llX.\n",my_addr); //DBG

	lseek(inputfd,my_addr,SEEK_SET); 

	// read whole page (4KB)
	if ( (status=read(inputfd,&tmp_small_page,I386_VPT_MAGIC_BS)) == -1 ) {
                errormsg=strerror(errno);
                fprintf(stderr, "i386_VPT_test: read() in %s: %s.\n",inputfile,errormsg);
                return 99;
        }
	
	for (mi=0;mi < (I386_VPT_MAGIC_BS/I386_POINTER_SIZE);mi++) { //page
		
		if (tmp_small_page[mi] != 0) {
			rcounter++; //found some record

			// in i386 in 4KB and 2MB modes in PDPE bits number 7 and 8 
			// are marked as MBZ (must be zero)
			// in 4MB mode, bit 7 must be 1
			// Pretty weak, but still better than nothing

			if (((tmp_small_page[mi] & 0x180) == 0x100)) { // only bit nr. 8 is set 
				ecounter++; //one more error
			}

			if (ecounter > I386_VPT_MAGIC_ERR) { 
//				dbgprint(" addr: 0x%llX mi=%d, ecounter: %d, rcounter: %d\n",my_addr,mi,ecounter,rcounter);
				return 1; //tolerate small amount of errors
			}
			
		}
	}

	//Another checks
	if ((rcounter == 0) || (ecounter > rcounter )) { 
//		dbgprint(" addr: 0x%llX mi=%d, ecounter: %d, rcounter: %d\n",my_addr,mi,ecounter,rcounter);
		return 2; // too few records
	
	}

	// All ok
//	dbgprint("PDPE at: 0x%llX mi=%d, ecounter: %d, rcounter: %d\n",my_addr,mi,ecounter,rcounter);
	return 0;
}

/*
 * Function for verifying if given address contains system VPT 
 * (resp. Page Map Level 4)
 *
 * Verification is done by simple checking
 * if first VPT_MAGIC_ZEROES bytes contain only zeroes and 
 * if flags for first three records beyond 0xC0000000
 * (beginning of kernel address space) are same.
 * This should usually be true, but it is just heuristic, 
 * and can fail.
 *
 *  x86_64_VPT_test (physical address to verify )
 *  return 0 on success, otherwise error > 0
 *
 * */

int x86_64_VPT_test(u_int64_t my_addr){

	int mi=0,ecounter=0,rcounter=0; //error counter
	u_int64_t tmp_small_page[X86_64_VPT_MAGIC_BS/X86_64_POINTER_SIZE];

//	dbgprint("x86_64_VPT_test: trying  0x%llX.\n",my_addr); //DBG

	lseek(inputfd,my_addr,SEEK_SET); 

	// read whole page (4KB)
	if ( (status=read(inputfd,&tmp_small_page,X86_64_VPT_MAGIC_BS)) == -1 ) {
                errormsg=strerror(errno);
                fprintf(stderr, "i386_VPT_test: read() in %s: %s.\n",inputfile,errormsg);
                return 99;
        }
	
	// there should be a lot of zeroes... add heuristic for this?

	for (mi=0;mi < (X86_64_VPT_MAGIC_BS/X86_64_POINTER_SIZE);mi++) { //page
		
		if (tmp_small_page[mi] != 0) { // TODO mask by special bits?
			rcounter++; //found some record

			// in x86_64 in all three modes (4KB, 2MB, 1GB) bits number 7 and 8
			// are marked as MBZ (must be zero)
			if ((tmp_small_page[mi] & 0x180) != 0) {
				//exit immediately?
				ecounter++; //one more error
			} else {	
				//only bits 51:12 contain address of PDPE
				if (x86_64_looks_like_page_table(tmp_small_page[mi] & 0x000FFFFFFFFFF000LLU) != 0) {
					ecounter++;//one more error
				}
			}

			if (ecounter > X86_64_VPT_MAGIC_ERR) {
//				dbgprint(" addr: 0x%llX mi=%d, ecounter: %d, rcounter: %d\n",my_addr,mi,ecounter,rcounter);
				return 1; //tolerate small amount of errors
			}
			
		}
	}

	//Another checks
	if ((rcounter == 0) || (ecounter > rcounter )) { 
//		dbgprint(" addr: 0x%llX mi=%d, ecounter: %d, rcounter: %d\n",my_addr,mi,ecounter,rcounter);
		return 2; // too few records
	
	}

	// All ok
//	dbgprint("Found VPT at: 0x%llX mi=%d, ecounter: %d, rcounter: %d\n",my_addr,mi,ecounter,rcounter);
	return 0;
}

/*	fprintf(stderr, "x86_64_VPT_test: trying  0x%llX.\n",my_addr); //DBG

	lseek(inputfd,my_addr,SEEK_SET); 

	if ( (status=read(inputfd,&tmp_small_page,X86_64_VPT_MAGIC_BS)) == -1 ) {
                errormsg=strerror(errno);
                fprintf(stderr, "i386_VPT_test: read() in %s: %s.\n",inputfile,errormsg);
                return 99;
        }
	
	for (mi=0;mi < X86_64_VPT_MAGIC_ZEROES;mi++) { //first 
		if (tmp_small_page[mi] != 0) {
			if (++ecounter > X86_64_VPT_MAGIC_ZEROES_ERR) {
//				fprintf(stderr,"i386_lin_VPT_test: addr: 0x%x mi=%d, ecounter: %d\n",my_addr,mi,ecounter); // DBG
				return 1; //tolerate small amount of errors
			}
		}
	}
//	fprintf(stderr,"i386_lin_VPT_test: addr: 0x%x ,zeroes, mi=%d, err=%d\n",my_addr,mi,ecounter); // DBG

	//check if structure *SEEMS* to be regular.
	// (!) NOTE: This is just simple heuristic, success not guaranteed
	//see documentation for more information
	//0x6b ~ 01101011 

	ecounter=0;
//	dbgprint("i386_lin_VPT_test:: dbg :mi: %d: %X %X %X\n",mi,tmp_small_page[mi],tmp_small_page[mi+4],tmp_small_page[mi+8]); //DBG 

	for (;mi < (X86_64_VPT_MAGIC_ZEROES+X86_64_VPT_MAGIC_ENTRIES_CHECK);mi=mi+image1.pointer_size) { 
		// test entries, (only first byte from each pointer) avoid trivial cases
		if (((tmp_small_page[mi] & 0x6b) != (tmp_small_page[mi+image1.pointer_size] & 0x6b)) || (tmp_small_page[mi] == 0)||(tmp_small_page[mi] == 0xFFFFFFFF)) {
			if (++ecounter > X86_64_VPT_MAGIC_ENTRIES_ERR) {
//				dbgprint("i386_lin_VPT_test: addr: 0x%x ,entries, mi=%d\n",my_addr,mi); // DBG
				return 2; //tolerate small amount of errors
			}
		}
	}
	fprintf(stderr,"x86_64_VPT_test: addr: 0x%llX ,entries, mi=%d, err=%d\n",my_addr,mi,ecounter); // DBG

	return 0; //all OK			
*/


#endif


/*
 * Function for verifying if given address contains VPT 
 * (resp. PaGeDirectory) of some process.
 *
 * Verification is done by simple checking
 * if flags for first three records beyond 0xC0000000
 * (beginning of kernel address space) are same.
 * This should usually be true, as userspace PGD are 
 * cloned from kernel PGD, but it is just heuristic, 
 * and can be wrong.
 * 
 *  XXX rewrite
 *
 *  i3886_user_I386_VPT_test (linear address to verify )
 *  return 0 on success, otherwise error > 0
 *
 * */


int i386_user_VPT_test(u_int64_t my_addr){

	int mi=0;
	unsigned char tmp_small_page[I386_VPT_MAGIC_BS];

	if ( (status=i386_read_linear(0,my_addr,tmp_small_page,I386_VPT_MAGIC_BS,NULL)) != 0 ) {
                errormsg=strerror(errno);
                fprintf(stderr, "i386_user_I386_VPT_test: read() in %s: %s.\n",inputfile,errormsg);
                return 99;
        }
	
//	printf("i386_lin_user_i386_lin_I386_VPT_test: %lx\n",my_addr);

	mi=I386_VPT_MAGIC_ZEROES; 

	//check if structure seems to be regular.
	// (!) NOTE: This is just simple heuristic, success not guaranteed
	//see documentation for more information
	// TODO: Change this to same test as used for kernel VPT?


//	printf("dbg: %2X %2X %2X\n",tmp_small_page[mi]-0xFFFFFF00,tmp_small_page[mi+4]-0xFFFFFF00,tmp_small_page[mi+8]-0xFFFFFF00); 
//	printf("user_vpt_test: dbg :i: %d: %X %X %X\n",mi,tmp_small_page[mi],tmp_small_page[mi+4],tmp_small_page[mi+8]); 

	if ((tmp_small_page[mi] & 0x6f) != (tmp_small_page[mi+image1.pointer_size] & 0x6f)) {return 2;}
	if ((tmp_small_page[mi] & 0x6f) != (tmp_small_page[mi+image1.pointer_size*2] & 0x6f)) {return 3;}
	if (tmp_small_page[mi] == 0) {return 4;}

//	printf("user_vpt_test: dbg2: :%X,%d,%c\n",tmp_small_page[mi]),tmp_small_page[mi],tmp_small_page[mi];

	return 0; //all OK			
}

/*
 * Function for verifying if given address contains system VPT 
 * (resp. PaGeDirectory)
 *
 * Verification is done by simple checking
 * if first I386_VPT_MAGIC_ZEROES bytes contain only zeroes and 
 * if flags for first three records beyond 0xC0000000
 * (beginning of kernel address space) are same.
 * This should usually be true, but it is just heuristic, 
 * and can be wrong. XXX rewrite
 *
 *  i386_VPT_test (physical address to verify )
 *  return 0 on success, otherwise error > 0
 *
 * */

int i386_VPT_test(u_int64_t my_addr){

	int mi=0,ecounter=0; //error counter
	unsigned char tmp_small_page[I386_VPT_MAGIC_BS];

//	dbgprint("trying  0x%X.\n",my_addr); //DBG

	lseek(inputfd,my_addr,SEEK_SET); 

	if ( (status=read(inputfd,&tmp_small_page,I386_VPT_MAGIC_BS)) == -1 ) {
                errormsg=strerror(errno);
                fprintf(stderr, "i386_VPT_test: read() in %s: %s.\n",inputfile,errormsg);
                return 99;
        }
	
	for (mi=0;mi < I386_VPT_MAGIC_ZEROES;mi++) { //first 
		if (tmp_small_page[mi] != 0) {
			if (++ecounter > I386_VPT_MAGIC_ZEROES_ERR) {
//				fprintf(stderr,"i386_lin_I386_VPT_test: addr: 0x%X mi=%d, ecounter: %d\n",my_addr,mi,ecounter); // DBG
				return 1; //tolerate small amount of errors
			}
		}
	}
//	fprintf(stderr,"i386_lin_I386_VPT_test: addr: 0x%X ,zeroes, mi=%d, err=%d\n",my_addr,mi,ecounter); // DBG

	//check if structure *SEEMS* to be regular.
	// (!) NOTE: This is just simple heuristic, success not guaranteed
	//see documentation for more information
	//0x6b ~ 01101011 

//	dbgprint("2: addr: %d.\n",image1.pointer_size); //DBG

	ecounter=0;
//	dbgprint("i386_lin_I386_VPT_test:: dbg :mi: %d: %X %X %X\n",mi,tmp_small_page[mi],tmp_small_page[mi+4],tmp_small_page[mi+8]); //DBG 

	for (;mi < (I386_VPT_MAGIC_ZEROES+I386_VPT_MAGIC_ENTRIES_CHECK);mi=mi+image1.pointer_size) { 
		// test entries, (only first byte from each pointer) avoid trivial cases
		if (((tmp_small_page[mi] & 0x6b) != (tmp_small_page[mi+image1.pointer_size] & 0x6b)) || (tmp_small_page[mi] == 0)||(tmp_small_page[mi] == 0x7F)) {
			if (++ecounter > I386_VPT_MAGIC_ENTRIES_ERR) {
//				dbgprint("i386_lin_I386_VPT_test: addr: 0x%X ,entries, mi=%d\n",my_addr,mi); // DBG
				return 2; //tolerate small amount of errors
			}
		}
	}
//	dbgprint("addr: 0x%llX ,entries, mi=%d, err=%d\n",my_addr,mi,ecounter); // DBG

	return 0; //all OK			
}


/*
 * Function that tries to determine general information 
 * about analyzed image.
 *
 * for now, only system VPT is found.
 *
 * i386_determine()
 * return 0 on success, otherwise error > 0
 * Discovered information are saved in global structure image1.
 *
 * */

int i386_determine(){
	
	u_int64_t addr=I386_VPT_MAGIC_ADDR;
	
	// I realised, that I don't get too much information in 
	// this phase of analysis 

	printf("i386_determine:\n");
	if (image1.VPT_user_specified == 0 ) {	// if image address was not manually specified
		while (i386_VPT_test(addr) != 0) {
	
			addr += I386_VPT_MAGIC_BS; // 4KB
//			dbgprint("step, addr: 0x%llX\n",addr);

			if (addr > ( I386_VPT_MAGIC_ADDR + I386_VPT_MAGIC_BS * I386_VPT_MAGIC_COUNT) ) { 
				// we have tried too many pages, so something is wrong probably
				// further search will be useless
				printf("I386_VPT_address not found, terminating at: 0x%llX\n",addr);
				return 1;
			}
		}
	
	
		image1.VPT_address=addr; // found one! :)
		printf("VPT address found at: 0x%llX\n",image1.VPT_address);
	} else {
		if (image1.VPT_not_available == 0 ) {	// Use of PGD is not disabled	
			printf("User specified VPT address: 0x%llX\n",image1.VPT_address);
		} else {
			printf("Using hardcoded mappings instead of VPT. This may produce strange results!\n");
		}
	}
	
	return 0;  // 0 if all OK, otherwise error
}


