/*
 *	Functions for analysing x64 architecture systems
*/

#include "../../globals.h"
#include "magic.h"
/*
 * Some alchemy:
 */

/* 
 * Function that manages mapping from linear to physical address space.
 * 
 * Good overview is available in 24593.pdf (AMD documentation) at page 117+
 *
 * priv_block_size from lin_addr are read using provided PGD (kernel PGD by default) into out_data pointer.
 * dump_offset is output parameter, if NULL, ignored.
 * Otherwise physical offset translated from linear address is stored here.
 * (!) Please note, that read block can be larger than one page. 
 * In this case offset is valid only for first part of data.
 */

int x86_64_read_linear(u_int64_t PGD, u_int64_t lin_addr, void *out_data, u_int64_t priv_block_size, u_int64_t *dump_offset){

	//"PGD" is needed, because each process have it's own linear address space. Process 0 means kernel memory.
	// NOTE: Term PGD is used to refer first directory in hierarchy,
	//       on x64 architecture this term is officially used for one of the levels (2),
        //       as x64 is based on i386 architecture.

	//Find in VPT "physical" address, where data should be. (if problem, return 1)
	//Read block_size of that data into out_data (if problem, return 2)
	//return 0

	u_int64_t physical_offset = lin_addr;

	u_int64_t PML4E_offset = lin_addr;
	u_int64_t PDPE_offset = lin_addr;
	u_int64_t PDE_offset = lin_addr;
	u_int64_t PTE_offset = lin_addr;
	u_int64_t page_number = lin_addr;

	u_int64_t tmp_var = 0;

//	dbgprint("addr: 0x%llX",lin_addr); 

	if (PGD == 0) {	//special case
		PGD = image1.VPT_address;

	/* if VPT_address is not available,
	 * we can try "best effort", and use hard-coded 
	 * "direct mapping", usually used by OS kernels.
	 * (and OS specific)
	 * for linux on x86_64, mapping can be found in 
	 * Documentation/x86/x86_64/mm.txt
	 * parsing here:

0000000000000000 - 00007fffffffffff (=47 bits) user space
ffff800000000000 - ffff80ffffffffff (=40 bits) guard hole
ffff880000000000 - ffffc7ffffffffff (=64 TB) direct mapping of all phys. memory
ffffc80000000000 - ffffc8ffffffffff (=40 bits) hole
ffffc90000000000 - ffffe8ffffffffff (=45 bits) vmalloc/ioremap space
ffffe90000000000 - ffffe9ffffffffff (=40 bits) hole
ffffea0000000000 - ffffeaffffffffff (=40 bits) virtual memory map (1TB)
... unused hole ...
ffffffff80000000 - ffffffffa0000000 (=512 MB)  kernel text mapping, from phys 0
ffffffffa0000000 - fffffffffff00000 (=1536 MB) module mapping space

	 * so we can try to translate ffff880000000000 - ffffc7ffffffffff
	 * and ffffffff80000000 - ffffffffa0000000
  	 */  
		if (image1.VPT_not_available == 1) { 
			//whole memory direct mapping
			if ((lin_addr > 0xFFFF880000000000LLU) && 
			    (lin_addr < 0xFFFFC7FFFFFFFFFFLLU)) {
				physical_offset=lin_addr & 0x000000FFFFFFFFFFLLU;
				dbgprint("direct mapping FFFF88");
			}

			//kernel text direct mapping
			if ((lin_addr > 0xFFFFFFFF80000000LLU) && 
			    (lin_addr < 0xFFFFFFFFA0000000LLU)) {
				physical_offset=lin_addr & 0x000000000FFFFFFFLLU;
				dbgprint("direct mapping FFFFFFFF8");
			}

			// Read memory
			lseek(inputfd,physical_offset,SEEK_SET); //level 0 (physical page) 
			dbgprint("(level 0) Dirrect mapping: Reading 0x%llX from PO: 0x%llX\n",lin_addr,physical_offset);
			
			if (dump_offset != NULL) { //save physical offset to output variable
				*dump_offset=physical_offset; 
			}

			if ( (status=read(inputfd,out_data,priv_block_size)) == -1 ) {
        		        errormsg=strerror(errno);
               			fprintf(stderr, "read_liner: read() in %s: %s.\n",inputfile,errormsg);
	                	return 38;
			}
			
			return 0; //another job well done.

		}	

	} //else PGD is physical address of PGD of some process

	//long mode	

	// this part (PML4E & PDPE) is same for all page sizes in long mode
	// For extracting addresses I use bitwise and, for extracting offsets I have to shift bits.
	PML4E_offset = ( PML4E_offset & 0x0000FFFFFFFFFFFFLLU ) >> 39; //bits ( 47:39)
	PDPE_offset =  ( PDPE_offset  & 0x0000007FFFFFFFFFLLU ) >> 30; //bits ( 38:30)

//	dbgprint("MPL4E offset (level 4): 0x%llX",PML4E_offset);	

        lseek(inputfd,PGD+PML4E_offset*image1.pointer_size,SEEK_SET); //level 4 (PML4E)
//	dbgprint("(level 4) reading from : 0x%llX",PGD+PML4E_offset*image1.pointer_size);	

        if ( (status=read(inputfd,&page_number,image1.pointer_size)) == -1 ) {
                errormsg=strerror(errno);
                fprintf(stderr, "read_liner: read() in %s: %s.\n",inputfile,errormsg);
                return 37;
        }
//	dbgprint("page_number (level 4): 0x%llX",page_number);	

	//bits ( 51:12), but no shift. Offset will be added later
	page_number = page_number & 0x000FFFFFFFFFF000LLU; 

//	dbgprint("PDPE offset (level 3): 0x%llX",PDPE_offset);
//	dbgprint("(level 3) reading from : 0x%llX",page_number+PDPE_offset*image1.pointer_size);
	lseek(inputfd,page_number+PDPE_offset*image1.pointer_size,SEEK_SET); //level 3 (PDPE)

        if ( (status=read(inputfd,&page_number,image1.pointer_size)) == -1 ) {
                errormsg=strerror(errno);
                fprintf(stderr, "read_liner: read() in %s: %s.\n",inputfile,errormsg);
                return 38;
        }


	// TODO: process all necessary special bits, like swapped, etc...

	//only bit NUMBER 7 (ie. 8th bit). This is PDPE.PS bit.
	tmp_var = page_number & 0x0000000000000080LLU; 
//	dbgprint("page_number (level 3): 0x%llX",page_number);	

	if (tmp_var == 1) { //1GB pages are used

//		dbgprint("1GB pages used for this page.");
	        physical_offset = physical_offset & 0x000000003FFFFFFFLLU; // bits 29:0
	
		//bits ( 51:30), but no shift. Offset will be added later
		page_number = page_number & 0x000FFFFFC0000000LLU; 


		lseek(inputfd,page_number+physical_offset,SEEK_SET); //level 0 (physical page) 
//		dbgprint("(level 0) Reading 0x%llX from PO: 0x%llX\n",lin_addr,page_number+physical_offset);

                if (dump_offset != NULL) { //save physical offset to output variable
                        *dump_offset=page_number+physical_offset;
                }

		//If we want to read more than 1 page:
		// (use recursion;)
		
		// offset + block_size must be less then 1GB
		if (priv_block_size > (GB_1-physical_offset)){ 
//			dbgprint("x86_64_read_linear: page too small for reading... processing another one.");
			//recursion	
			tmp_var = GB_1-physical_offset;

			//Process rest of request recursively... beware about effectives
			x86_64_read_linear(PGD,lin_addr+tmp_var,out_data+tmp_var,priv_block_size-tmp_var,NULL); 
		} else{
			tmp_var=priv_block_size;
		}

		if ( (status=read(inputfd,out_data,tmp_var)) == -1 ) {
        	        errormsg=strerror(errno);
               		fprintf(stderr, "read_liner: read() in %s: %s.\n",inputfile,errormsg);
	                return 38;
		}


	} else { // 2MB or 4KB pages are used

		//bits ( 51:12), but no shift. Offset will be added later
		page_number = page_number & 0x000FFFFFFFFFF000LLU; 

		PDE_offset = ( PDE_offset & 0x000000003FE00000LLU ) >> 21; //bits ( 29:21)

//		dbgprint("test page number (pre level 2): 0x%llX",page_number);
//		dbgprint("PDE offset (level 2): 0x%llX",PDE_offset);
//	 	dbgprint("(level 2) reading from : 0x%llX",page_number+PDE_offset*image1.pointer_size);       
		lseek(inputfd,page_number+PDE_offset*image1.pointer_size,SEEK_SET); //level 2 (PDE)

        	if ( (status=read(inputfd,&page_number,image1.pointer_size)) == -1 ) {
	                errormsg=strerror(errno);
	                fprintf(stderr, "read_liner: read() in %s: %s.\n",inputfile,errormsg);
	                return 39;
	        }

//		dbgprint("page_number (level 2): 0x%llX",page_number);	
		//only bit NUMBER 7 (ie. 8th bit). This is PDE.PS bit.
		tmp_var = page_number & 0x0000000000000080LLU; 
	
		if (tmp_var == 0) { // 4KB pages are used
//			dbgprint("4KB pages used for this page.");

			PTE_offset = ( PTE_offset & 0x00000000001FF000LLU ) >> 12; //bits ( 20:12)
		        physical_offset = physical_offset & 0x00000FFF; // bits 11:0

//			dbgprint("(level 1)PTE_offset: 0x%llX",PTE_offset);
//			dbgprint("(level 1)Orig page_number: 0x%llX",page_number);
//			dbgprint("(level 1)physical_offset: 0x%llX",physical_offset);

			//bits ( 51:12), but no shift. Offset will be added later
			page_number = page_number & 0x000FFFFFFFFFF000LLU; 
//			dbgprint("(level 1)page_number: 0x%llX",page_number);

			lseek(inputfd,page_number+PTE_offset*image1.pointer_size,SEEK_SET); //level 1 (PTE)
//			dbgprint("(level 1)Reading from PO: 0x%llX",page_number+PTE_offset*image1.pointer_size);


			if ( (status=read(inputfd,&page_number,image1.pointer_size)) == -1 ) {
	        	        errormsg=strerror(errno);
	                	fprintf(stderr, "read_liner: read() in %s: %s.\n",inputfile,errormsg);
				return 39;
			}

			//bits ( 51:12), but no shift. Offset will be added later
			page_number = page_number & 0x000FFFFFFFFFF000LLU; 
//			dbgprint("(level 0)page_number: 0x%llX",page_number);

//			dbgprint("(level 0)Reading from PO: 0x%llX",page_number+physical_offset);
			lseek(inputfd,page_number+physical_offset,SEEK_SET); //level 0 (physical page)
			if (dump_offset != NULL) { //save physical offset to output variable
				*dump_offset=page_number+physical_offset; 
//				dbgprint("(level 0)dump_offset: 0x%llX",*dump_offset);
			}

			//If we want to read more than 1 page:
			// (use recursion;)
		
			// offset + block_size must be less then 4KB
			if (priv_block_size > (KB_4-physical_offset)){ 
//				dbgprint("page too small for reading... processing another one.");
				//recursion	
				tmp_var = KB_4-physical_offset;

				//Process rest of request recursively... beware about effectives
				x86_64_read_linear(PGD,lin_addr+tmp_var,out_data+tmp_var,priv_block_size-tmp_var,NULL); 

			} else{
				tmp_var=priv_block_size;
			}

			if ( (status=read(inputfd,out_data,tmp_var)) == -1 ) {
        		        errormsg=strerror(errno);
                		fprintf(stderr, "read_liner: read() in %s: %s.\n",inputfile,errormsg);
//				dbgprint("(level 0)read itself failed.\n");

		                return 38;
			}
//			dbgprint("(level 0)read itself ok, status=%d\n",status);



		} else { // 2MB pages are used
//			dbgprint("2MB pages used for this page."); 

		        physical_offset = physical_offset & 0x00000000001FFFFFLLU; // bits 20:0
//			dbgprint("(level 0)Orig page_number: 0x%llX",page_number);

			//bits ( 51:21), but no shift. Offset will be added later
			page_number = page_number & 0x000FFFFFFFE00000LLU; 
//			dbgprint("(level 0) used page_number: 0x%llX",page_number);
//			dbgprint("(level 0) physical_offset: 0x%llX",physical_offset);

			lseek(inputfd,page_number+physical_offset,SEEK_SET); //level 0 (physical page) 
//			dbgprint("(level 0) Reading 0x%llX from PO: 0x%llX",lin_addr,page_number+physical_offset);

//			dbgprint("dump_offset: %p",dump_offset);
			if (dump_offset != NULL) { //save physical offset to output variable
				*dump_offset=page_number+physical_offset; 
//				dbgprint("(level 0)dump_offset: 0x%llX",*dump_offset);
			}

			//If we want to read more than 1 page:
			// (use recursion;)
		
			// offset + block_size must be less then 2MB
			if (priv_block_size > (MB_2-physical_offset)){ 
//				dbgprint("page too small for reading... processing another one.\n");
				//recursion	
				tmp_var = MB_2-physical_offset;

				//Process rest of request recursively... beware about effectives
				x86_64_read_linear(PGD,lin_addr+tmp_var,out_data+tmp_var,priv_block_size-tmp_var,NULL); 

			} else{
				tmp_var=priv_block_size;
			}

			if ( (status=read(inputfd,out_data,tmp_var)) == -1 ) {
        		        errormsg=strerror(errno);
                		fprintf(stderr, "read_liner: read() in %s: %s.\n",inputfile,errormsg);
//				dbgprint("(level 0)read itself failed.\n");

		                return 38;
			}
//			dbgprint("(level 0)read itself ok, status=%d\n",status);


		}

	}

        return 0;

}


/*
 * Function for verifying if given address contains VPT 
 * (resp. PaGeDirectory) of some process.
 */
/*
int x86_64_user_VPT_test(u_int64_t my_addr){

	dbgprint("TODO: my_addr:0x%llX",my_addr);
	return 0;
}
*/

/*
 * Function for verifying if given address contains
 * PDPE from page tables.
 *
 * Verification is done by simple checking
 * if all entries contain only zeroes or 
 * if they contain entry with certain valid bits.
 * This should usually be true, but it is just heuristic, 
 * and can fail.
 *
 * x86_64_looks_like_page_table (physical address to verify )
 * return 0 on success, otherwise error > 0
 *
 */


int x86_64_looks_like_page_table(u_int64_t my_addr){

	int mi=0,ecounter=0,rcounter=0; //error counter
	u_int64_t tmp_small_page[X86_64_VPT_MAGIC_BS/X86_64_POINTER_SIZE]; //4KB

//	dbgprint("x86_64_VPT_test: trying  0x%llX.\n",my_addr); //DBG

	lseek(inputfd,my_addr,SEEK_SET); 

	// read whole page (4KB)
	if ( (status=read(inputfd,&tmp_small_page,X86_64_VPT_MAGIC_BS)) == -1 ) {
                errormsg=strerror(errno);
                fprintf(stderr, "i386_VPT_test: read() in %s: %s.\n",inputfile,errormsg);
                return 99;
        }
	
	for (mi=0;mi < (X86_64_VPT_MAGIC_BS/X86_64_POINTER_SIZE);mi++) { //page
		
		if (tmp_small_page[mi] != 0) { //TOODO mask by special bits?
			rcounter++; //found some record

			// in x86_64 in 4KB and 2MB modes in PDPE bits number 7 and 8 
			// are marked as MBZ (must be zero)
			// in 1GB mode, bit 7 must be 1
			// Pretty weak, but still better than nothing

			if (((tmp_small_page[mi] & 0x180) == 0x100)) { // only bit nr. 8 is set
				ecounter++; //one more error
			}

			if (ecounter > X86_64_VPT_MAGIC_ERR) { //TODO use different constant?
//				dbgprint(" addr: 0x%llX mi=%d, ecounter: %d, rcounter: %d\n",my_addr,mi,ecounter,rcounter);
				return 1; //tolerate small amount of errors
			}
			
		}
	}

	//Another checks
	if ((rcounter == 0) || (ecounter > rcounter )) { 
//		dbgprint(" addr: 0x%llX mi=%d, ecounter: %d, rcounter: %d\n",my_addr,mi,ecounter,rcounter);
		return 2; // too few records
	
	}

	// All ok
//	dbgprint("PDPE at: 0x%llX mi=%d, ecounter: %d, rcounter: %d\n",my_addr,mi,ecounter,rcounter);
	return 0;
}

/*
 * Function for verifying if given address contains system VPT 
 * (resp. Page Map Level 4)
 *
 * Verification is done by simple checking
 * if first VPT_MAGIC_ZEROES bytes contain only zeroes and 
 * if flags for first three records beyond 0xC0000000
 * (beginning of kernel address space) are same.
 * This should usually be true, but it is just heuristic, 
 * and can fail.
 *
 *  x86_64_VPT_test (physical address to verify )
 *  return 0 on success, otherwise error > 0
 *
 * */

int x86_64_VPT_test(u_int64_t my_addr){

	int mi=0,ecounter=0,rcounter=0; //error counter
	u_int64_t tmp_small_page[X86_64_VPT_MAGIC_BS/X86_64_POINTER_SIZE];

//	dbgprint("x86_64_VPT_test: trying  0x%llX.\n",my_addr); //DBG

	lseek(inputfd,my_addr,SEEK_SET); 

	// read whole page (4KB)
	if ( (status=read(inputfd,&tmp_small_page,X86_64_VPT_MAGIC_BS)) == -1 ) {
                errormsg=strerror(errno);
                fprintf(stderr, "i386_VPT_test: read() in %s: %s.\n",inputfile,errormsg);
                return 99;
        }
	
	// TODO there should be a lot of zeroes... add heuristic for this?

	for (mi=0;mi < (X86_64_VPT_MAGIC_BS/X86_64_POINTER_SIZE);mi++) { //page
		
		if (tmp_small_page[mi] != 0) { //TODO mask by special bits?
			rcounter++; //found some record

			// in x86_64 in all three modes (4KB, 2MB, 1GB) bits number 7 and 8
			// are marked as MBZ (must be zero)
			if ((tmp_small_page[mi] & 0x180) != 0) {
				ecounter++; //one more error
			} else {	
				//only bits 51:12 contain address of PDPE
				if (x86_64_looks_like_page_table(tmp_small_page[mi] & 0x000FFFFFFFFFF000LLU) != 0) {
					ecounter++;//one more error
				}
			}

			if (ecounter > X86_64_VPT_MAGIC_ERR) {
//				dbgprint(" addr: 0x%llX mi=%d, ecounter: %d, rcounter: %d\n",my_addr,mi,ecounter,rcounter);
				return 1; //tolerate small amount of errors
			}
			
		}
	}

	//Another checks
	if ((rcounter == 0) || (ecounter > rcounter )) { 
//		dbgprint(" addr: 0x%llX mi=%d, ecounter: %d, rcounter: %d\n",my_addr,mi,ecounter,rcounter);
		return 2; // too few records
	
	}

	// All ok
//	dbgprint("Found VPT at: 0x%llX mi=%d, ecounter: %d, rcounter: %d\n",my_addr,mi,ecounter,rcounter);
	return 0;
}





/*
 * Function that tries to determine general information 
 * about analyzed image.
 *
 * for now, only system VPT is found.
 *
 * x86_64_determine ()
 * return 0 on success, otherwise error > 0
 * Discovered information are saved in global structure image1.
 *
 */


int x86_64_determine(){
	
	u_int64_t addr=X86_64_VPT_MAGIC_ADDR;
	
	// I realised, that I don't get too much information in 
	// this phase of analysis 
	
	printf("X86_64_determine\n");
	
	if (image1.VPT_user_specified == 0 ) {	// if image address was not manually specified
		while (x86_64_VPT_test(addr) != 0) {
	
			addr += X86_64_VPT_MAGIC_BS; // 4KB
//			dbgprint("step, addr: 0x%llX\n",addr);

			if (addr > ( X86_64_VPT_MAGIC_ADDR + X86_64_VPT_MAGIC_BS * X86_64_VPT_MAGIC_COUNT) ) { 
				// we have tried too many pages, so something is wrong probably
				// further search will be useless
				printf("X86_64 VPT_address not found, terminating at: 0x%llX\n",addr);
				return 1;
			}
		}
		
		image1.VPT_address=addr; // found one! :)
		printf("VPT address found at: 0x%llX\n",image1.VPT_address);
	} else {
		if (image1.VPT_not_available == 0 ) {	// Use of PGD is not disabled	
			printf("User specified VPT address: 0x%llX\n",image1.VPT_address);
		} else {
			printf("Using hardcoded mappings instead of VPT. This may produce strange results!\n");
		}
	}

	return 0;  // 0 if all OK, otherwise error
}

